# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2019  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""About dialog implementation"""

from utils.pixmapcache import getPixmap
from utils.globals import GlobalData
from utils.versions import getComponentInfo
from flowui.cml import CMLVersion
from .qt import (QDialog, QVBoxLayout, QHBoxLayout, QLabel, QTabWidget,
                 QTextBrowser, QDialogButtonBox, QSizePolicy, Qt)


class AboutDialog(QDialog):

    """Codimension about dialog"""

    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self.__createLayout()
        self.setWindowTitle("About Codimension")

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(640, 480)
        self.setSizeGripEnabled(True)

        vboxLayout = QVBoxLayout(self)
        hboxLayout = QHBoxLayout()
        iconLabel = QLabel()
        iconLabel.setPixmap(getPixmap("logo-48x48.png"))
        iconLabel.setScaledContents(False)
        hboxLayout.addWidget(iconLabel)
        versionLabel = QLabel(
            "<b>Codimension IDE version " +
            str(GlobalData().version) + "<br>"
            "CML version " +
            str(CMLVersion.VERSION) +
            "</b><p>Copyright (c) Sergey Satskiy 2010-2019</p>")
        versionLabel.setSizePolicy(QSizePolicy.Expanding,
                                   QSizePolicy.Expanding)
        versionLabel.setFixedHeight(versionLabel.minimumSizeHint().height())
        versionLabel.setAlignment(Qt.AlignCenter)
        hboxLayout.addWidget(versionLabel)
        vboxLayout.addLayout(hboxLayout)

        tabWidget = QTabWidget(self)
        tabWidget.setFocusPolicy(Qt.NoFocus)

        description = self.__createDescription()
        tabWidget.addTab(description, "About")
        versioning = self.__createVersioning()
        tabWidget.addTab(versioning, "Versions and licenses")
        authors = self.__createAuthors()
        tabWidget.addTab(authors, "Contributors")
        vboxLayout.addWidget(tabWidget)

        # Button at the bottom
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Ok)
        buttonBox.accepted.connect(self.close)
        buttonBox.rejected.connect(self.close)
        vboxLayout.addWidget(buttonBox)

    @staticmethod
    def __createDescription():
        """Creates the description browser"""
        browser = QTextBrowser()
        browser.setHtml("""
<p align="justify">
Codimension is a free experimental Python IDE.
</p>
<p align="justify">
Codimension aims to provide an integrated system for:
<ul>
<li>traditional text-based code editing, and</li>
<li>diagram-based code analysis (e.g. imports and classes diagrams,
    profiling results etc).</li>
</ul>
</p>

<p align="justify">
Many Python developers will find codimension useful as-is, even though not all
of its features have been implemented yet.
</p>
<p align="justify">
More information could be found on <a href='http://codimension.org'>
Codimension home page</a>.<br> Happy coding with Codimension!
</p>""")
        browser.setOpenExternalLinks(True)
        return browser

    @staticmethod
    def __createVersioning():
        """Creates the versioning section"""
        components = getComponentInfo()
        text = '<p>The major Codimension components are listed below:</p><ul>'
        for (prettyName, version, homeURL,
             patched, licenseName, licenseLink, localPath) in components:
            if homeURL is not None:
                text += "<li><a href='" + homeURL + "'>" + \
                        prettyName + "</a><br/>Version: "
            else:
                text += '<li>' + prettyName + '<br/>Version: '

            needRest = False
            if version is None:
                text += 'not installed'
            elif version == '':
                text += 'n/a'
            elif version.lower() == 'not installed':
                text += 'not installed'
            else:
                text += str(version)
                needRest = True

            if needRest:
                if patched:
                    text += ' (patched for codimension)'

                if licenseLink is not None and licenseName is not None:
                    text += '<br/>'
                    text += 'License: '
                    if licenseLink.startswith('http'):
                        text += "<a href='" + licenseLink + "'>" + \
                                licenseName + "</a>"
                    else:
                        text += licenseName + ' (' + licenseLink + ')'

                if localPath:
                    text += '<br/>Local path: ' + localPath

            text += '<br/></li>'
        text += '<li>Some icons (svg and png) are made by ' \
                '<a href="https://www.flaticon.com/authors/dave-gandy">' \
                'Dave Gandy</a> from ' \
                '<a href="https://www.flaticon.com/">www.flaticon.com</a>' \
                '<br/></li>'
        text += '</ul>'
        browser = QTextBrowser()
        browser.setHtml(text)
        browser.setOpenExternalLinks(True)
        return browser

    def __createAuthors(self):
        """Creates the authors section"""
        addr1 = "<a href='mailto:sergey.satskiy@gmail.com'>" \
                "sergey.satskiy@gmail.com</a>"
        addr2 = "<a href='mailto:isloginov@gmail.com'>isloginov@gmail.com</a>"
        addr3 = "<a href='mailto:him@revl.org'>him@revl.org</a>"
        addr4 = "<a href='mailto:david@dm9606.com'>david@dm9606.com</a>"
        addr5 = "<a href='mailto:fukanchik@gmail.com'>fukanchik@gmail.com</a>"
        text = "<p>Author: Sergey Satskiy &lt;" + addr1 + "&gt;</p>" \
               "<p>Packaging and CI (travis): Sergey Fukanchik &lt;" + \
               addr5 + "&gt;</p>" \
               "<p>Packaging: Ilya Loginov &lt;" + addr2 + "&gt;</p>" \
               "<p>Discussions, ideas, testing: David McElhany &lt;" + \
               addr4 + "&gt;</p>" \
               "<p>Initial packaging and logo: Dmitry Kazimirov &lt;" + \
               addr3 + "&gt;</p>" \
               "<p align='justify'>Special thanks: Detlev Offenbach - " \
               "the author of the Eric Python IDE." \
               " Codimension borrows some code and some ideas from Eric 4.</p>"
        browser = QTextBrowser()
        browser.setHtml(text)
        browser.setOpenExternalLinks(True)
        return browser



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Functions browser model"""

from os.path import basename, realpath
from utils.project import CodimensionProject
from utils.fileutils import isPythonFile
from utils.settings import Settings
from .viewitems import TreeViewFunctionItem
from .browsermodelbase import BrowserModelBase


class FunctionsBrowserModel(BrowserModelBase):

    """Class implementing the project browser model"""

    def __init__(self, parent=None):
        BrowserModelBase.__init__(self, ["Name", "File name", "Line"], parent)
        self.setTooltips(Settings()['functionsTooltips'])
        self.globalData.project.sigProjectChanged.connect(
            self.__onProjectChanged)

    def __populateModel(self):
        """Populates the project browser model"""
        self.clear()
        project = self.globalData.project
        cache = self.globalData.briefModinfoCache
        for fname in project.filesList:
            if isPythonFile(fname):
                info = cache.get(fname)
                for func in info.functions:
                    item = TreeViewFunctionItem(self.rootItem, func)
                    item.appendData([basename(fname), func.line])
                    item.setPath(fname)
                    self.rootItem.appendChild(item)

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what == CodimensionProject.CompleteProject:
            self.__populateModel()

    def onFSChanged(self, addedPythonFiles, deletedPythonFiles):
        """Triggered when some files appeared or disappeared"""
        needUpdate = False
        itemsToDelete = []
        for path in deletedPythonFiles:
            for item in self.rootItem.childItems:
                if realpath(path) == realpath(item.getPath()):
                    itemsToDelete.append(item)

        for item in itemsToDelete:
            needUpdate = True
            self.removeTreeItem(item)

        for path in addedPythonFiles:
            try:
                info = self.globalData.briefModinfoCache.get(path)
            except:
                # It could be that a file was created and deleted straight
                # away. In this case the cache will generate an exception.
                continue
            for funcObj in info.functions:
                needUpdate = True
                newItem = TreeViewFunctionItem(self.rootItem, funcObj)
                newItem.appendData([basename(path), funcObj.line])
                newItem.setPath(path)
                self.addTreeItem(self.rootItem, newItem)
        return needUpdate

    def onFileUpdated(self, fileName):
        """Triggered when a file was updated"""
        # Here: python file which belongs to the project
        info = self.globalData.briefModinfoCache.get(fileName)

        existingFunctions = []
        itemsToRemove = []
        needUpdate = False

        # For all root items
        path = realpath(fileName)
        for treeItem in self.rootItem.childItems:
            if realpath(treeItem.getPath()) != path:
                continue

            # Item belongs to the modified file
            name = treeItem.sourceObj.name
            found = False
            for func in info.functions:
                if func.name == name:
                    found = True
                    existingFunctions.append(name)
                    treeItem.updateData(func)
                    treeItem.setData(2, func.line)
                    self.signalItemUpdated(treeItem)
                    self.updateSingleFuncItem(treeItem, func)
                    break
            if not found:
                itemsToRemove.append(treeItem)

        for item in itemsToRemove:
            needUpdate = True
            self.removeTreeItem(item)

        # Add those which have been introduced
        for item in info.functions:
            if item.name not in existingFunctions:
                needUpdate = True
                newItem = TreeViewFunctionItem(self.rootItem, item)
                newItem.appendData([basename(fileName), item.line])
                newItem.setPath(fileName)
                self.addTreeItem(self.rootItem, newItem)
        return needUpdate



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2015-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Various items used to represent a control flow on a virtual canvas"""

# pylint: disable=C0305
# pylint: disable=C0302
# pylint: disable=W0702
# pylint: disable=R0913

from sys import maxsize
from html import escape
from ui.qt import Qt, QPen, QBrush, QGraphicsRectItem, QGraphicsItem
from .auxitems import BadgeItem, Connector, HSpacerCell, VSpacerCell, SpacerCell
from .cellelement import CellElement
from .routines import distance, getNoCellCommentBoxPath
from .cml import CMLVersion
from .colormixin import ColorMixin
from .textmixin import TextMixin
from .iconmixin import IconMixin


class ScopeHSideEdge(HSpacerCell):

    """Reserves some space for the scope horizontal edge"""

    def __init__(self, ref, canvas, x, y):
        HSpacerCell.__init__(self, ref, canvas, x, y,
                             width=canvas.settings.scopeRectRadius +
                             canvas.settings.hCellPadding)
        self.kind = CellElement.SCOPE_H_SIDE_EDGE


class ScopeVSideEdge(VSpacerCell):

    """Reserves some space for the scope vertical edge"""

    def __init__(self, ref, canvas, x, y):
        VSpacerCell.__init__(self, ref, canvas, x, y,
                             height=canvas.settings.scopeRectRadius +
                             canvas.settings.vCellPadding)
        self.kind = CellElement.SCOPE_V_SIDE_EDGE


class ScopeSpacer(SpacerCell):

    """Reserves some space for the scope corner"""

    def __init__(self, ref, canvas, x, y):
        SpacerCell.__init__(self, ref, canvas, x, y,
                            width=canvas.settings.scopeRectRadius +
                            canvas.settings.hCellPadding,
                            height=canvas.settings.scopeRectRadius +
                            canvas.settings.vCellPadding)
        self.kind = CellElement.SCOPE_CORNER_EDGE


class ScopeCellElement(CellElement, TextMixin, ColorMixin, IconMixin,
                       QGraphicsRectItem):

    """Base class for the scope items"""

    TOP_LEFT = 0
    DECLARATION = 1
    COMMENT = 2
    DOCSTRING = 3

    def __init__(self, ref, canvas, x, y, subKind,
                 bgColor, fgColor, borderColor):
        isDocstring = subKind == ScopeCellElement.DOCSTRING
        if subKind == ScopeCellElement.DOCSTRING:
            bgColor = canvas.settings.docstringBGColor
            fgColor = canvas.settings.docstringFGColor
            # Border color is borrowed from the scope for docstrings

        commentIconFileName = None
        if subKind == ScopeCellElement.COMMENT:
            if canvas.settings.hidecomments:
                commentIconFileName = 'hiddencomment.svg'

        CellElement.__init__(self, ref, canvas, x, y)
        TextMixin.__init__(self)
        ColorMixin.__init__(self, ref, bgColor, fgColor, borderColor,
                            isDocstring=isDocstring)
        IconMixin.__init__(self, canvas, commentIconFileName)
        QGraphicsRectItem.__init__(self)

        self.subKind = subKind
        self.badgeItem = None
        self.__docBadge = None
        self.__navBarUpdate = None
        self._connector = None
        self._topHalfConnector = None
        self.scene = None
        self.__sideCommentPath = None

        # Will be initialized only for the TOP_LEFT item of the
        # ELSE_SCOPE, EXCEPT_SCOPE and FINALLY_SCOPE
        # It points to TRY, FOR and WHILE approprietely
        self.leaderRef = None

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def __renderTopLeft(self):
        """Renders the top left corner of the scope"""
        s = self.canvas.settings
        self.minHeight = s.scopeRectRadius + s.vCellPadding
        self.minWidth = s.scopeRectRadius + s.hCellPadding

    def __renderDeclaration(self):
        """Renders the scope declaration"""
        # The declaration location uses a bit of the top cell space
        # to make the view more compact
        s = self.canvas.settings
        self.setupText(self)

        badgeItem = self.canvas.cells[
            self.addr[1] - 1][self.addr[0] - 1].badgeItem

        self.minHeight = self.textRect.height() + \
                         2 * s.vHeaderPadding - s.scopeRectRadius
        headerRectWidth = self.textRect.width()
        if badgeItem:
            headerRectWidth = max(headerRectWidth, badgeItem.width)
        self.minWidth = headerRectWidth + s.hHeaderPadding - s.scopeRectRadius
        if badgeItem:
            if badgeItem.withinHeader():
                self.minWidth = badgeItem.width + \
                                s.hHeaderPadding - s.scopeRectRadius
        if hasattr(self.ref, "sideComment"):
            if self.ref.sideComment:
                self.minHeight += 2 * s.vTextPadding
                self.minWidth += s.hCellPadding
            else:
                self.minHeight += s.vTextPadding
                self.minWidth += s.hHeaderPadding
        else:
            self.minWidth += s.hHeaderPadding
        self.minWidth = max(self.minWidth, s.minWidth)

    def __renderComment(self):
        """Renders the scope declaration"""
        s = self.canvas.settings
        self.setupText(self, customText=self.getSideComment(),
                       customReplacement='')
        if s.hidecomments:
            # Visually the icon looks a bit too big so reduce the size to 80%
            self.iconItem.setIconHeight(self.iconItem.iconHeight() * 0.8)

            self.minHeight = self.iconItem.iconHeight() + \
                2 * (s.vHeaderPadding + s.vHiddenCommentPadding) - \
                s.scopeRectRadius
            self.minWidth = s.hCellPadding + self.iconItem.iconWidth() + \
                2 * s.hHiddenCommentPadding + s.hHeaderPadding - \
                s.scopeRectRadius
        else:
            self.minHeight = self.textRect.height() + \
                2 * (s.vHeaderPadding + s.vTextPadding) - \
                s.scopeRectRadius
            self.minWidth = s.hCellPadding + s.hTextPadding + \
                self.textRect.width() + s.hTextPadding + \
                s.hHeaderPadding - s.scopeRectRadius

    def __renderDocstring(self):
        """Renders the scope docstring"""
        s = self.canvas.settings
        self.setupText(self, customText=self.ref.docstring.getDisplayValue(),
                       customReplacement='')

        if s.hidedocstrings:
            self.__docBadge = BadgeItem(self, 'doc')
            self.__docBadge.setToolTip('<pre>' + escape(self.text) + '</pre>')
            self.minHeight = self.__docBadge.height + 2 * (s.selectPenWidth - 1)
            self.minWidth = 2 * (s.hHeaderPadding - s.scopeRectRadius)
        else:
            self.minHeight = self.textRect.height() + 2 * s.vHeaderPadding
            self.minWidth = self.textRect.width() + 2 * (s.hHeaderPadding -
                                                         s.scopeRectRadius)

    def renderCell(self):
        """Provides rendering for the scope elements"""
        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.__renderTopLeft()
        elif self.subKind == ScopeCellElement.DECLARATION:
            self.__renderDeclaration()
        elif self.subKind == ScopeCellElement.COMMENT:
            self.__renderComment()
        elif self.subKind == ScopeCellElement.DOCSTRING:
            self.__renderDocstring()
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def __followLoop(self):
        """Used to detect if an 'else' scope is for a loop"""
        row = self.canvas.addr[1]
        column = self.canvas.addr[0] - 1
        cells = self.canvas.canvas.cells
        try:
            if cells[row][column].kind == CellElement.VCANVAS:
                return cells[row][column].cells[0][0].kind \
                    in [CellElement.FOR_SCOPE, CellElement.WHILE_SCOPE]
            return False
        except:
            return False

    def __needConnector(self):
        """True if a connector is required"""
        if self.kind in [CellElement.FOR_SCOPE, CellElement.DECOR_SCOPE,
                         CellElement.WHILE_SCOPE, CellElement.FUNC_SCOPE,
                         CellElement.CLASS_SCOPE, CellElement.WITH_SCOPE,
                         CellElement.FINALLY_SCOPE, CellElement.TRY_SCOPE]:
            return True
        if self.kind == CellElement.ELSE_SCOPE:
            return self.statement == ElseScopeCell.TRY_STATEMENT
        return False

    def __needTopHalfConnector(self):
        """True if a half of a connector is needed"""
        if self.kind in [CellElement.ELSE_SCOPE, CellElement.EXCEPT_SCOPE]:
            try:
                parentCanvas = self.canvas.canvas
                cellToTheTop = parentCanvas.cells[
                    self.canvas.addr[1] - 1][self.canvas.addr[0]]
                return cellToTheTop.needConnector
            except:
                pass
        return False

    def __drawTopLeft(self):
        """Draws the top left element of a scope"""
        s = self.canvas.settings

        # Draw connector if needed
        if self.__needConnector() and self._connector is None:
            self._connector = Connector(
                self.canvas, self.baseX + s.mainLine,
                self.baseY, self.baseX + s.mainLine,
                self.baseY + self.canvas.height)
            self.scene.addItem(self._connector)
        if self.__needTopHalfConnector() and self._topHalfConnector is None:
            self._topHalfConnector = Connector(
                self.canvas, self.baseX + s.mainLine, self.baseY,
                self.baseX + s.mainLine, self.baseY + self.canvas.height / 2)
            self.scene.addItem(self._topHalfConnector)

        # Draw the scope rounded rectangle when we see the top left corner
        penWidth = s.selectPenWidth - 1
        self.setRect(
            self.baseX + s.hCellPadding - penWidth,
            self.baseY + s.vCellPadding - penWidth,
            self.canvas.minWidth - 2 * s.hCellPadding + 2 * penWidth,
            self.canvas.minHeight - 2 * s.vCellPadding + 2 * penWidth)
        self.scene.addItem(self)
        self.canvas.scopeRectangle = self
        if self.badgeItem:
            if self.badgeItem.withinHeader():
                headerHeight = self.canvas.cells[
                    self.addr[1] + 1][self.addr[0]].height
                fullHeight = headerHeight + s.scopeRectRadius
                self.badgeItem.moveTo(
                    self.baseX + s.hCellPadding + s.badgeShift,
                    self.baseY + s.vCellPadding + fullHeight / 2 -
                    self.badgeItem.height / 2)
            else:
                self.badgeItem.moveTo(
                    self.baseX + s.hCellPadding + s.badgeShift,
                    self.baseY + s.vCellPadding - self.badgeItem.height / 2)
            self.scene.addItem(self.badgeItem)
        # Draw a horizontal connector if needed
        if self._connector is None:
            if self.kind == CellElement.EXCEPT_SCOPE or (
                    self.kind == CellElement.ELSE_SCOPE and
                    self.__followLoop()):
                parentCanvas = self.canvas.canvas
                cellToTheLeft = parentCanvas.cells[
                    self.canvas.addr[1]][self.canvas.addr[0] - 1]
                self._connector = Connector(
                    self.canvas,
                    cellToTheLeft.baseX + cellToTheLeft.minWidth -
                    s.hCellPadding + s.boxLineWidth,
                    self.baseY + 2 * s.vCellPadding,
                    self.baseX + s.hCellPadding - s.boxLineWidth,
                    self.baseY + 2 * s.vCellPadding)
                self._connector.penStyle = Qt.DotLine
                self.scene.addItem(self._connector)

        if hasattr(self.scene.parent(), "updateNavigationToolbar"):
            self.__navBarUpdate = self.scene.parent().updateNavigationToolbar
            self.setAcceptHoverEvents(True)

    def __drawDeclaration(self):
        """Draws the declaration item"""
        s = self.canvas.settings
        penWidth = s.selectPenWidth - 1
        self.setRect(
            self.baseX - s.scopeRectRadius - penWidth,
            self.baseY - s.scopeRectRadius - penWidth,
            self.canvas.minWidth - 2 * s.hCellPadding + 2 * penWidth,
            self.height + s.scopeRectRadius + penWidth)
        self.scene.addItem(self)

    def __drawComment(self):
        """Draws the comment item"""
        s = self.canvas.settings
        canvasTop = self.baseY - s.scopeRectRadius
        movedBaseX = self.canvas.baseX + self.canvas.minWidth - \
            self.width - s.scopeRectRadius - s.vHeaderPadding
        penWidth = s.selectPenWidth - 1
        if s.hidecomments:
            self.setRect(
                movedBaseX + s.hHeaderPadding - penWidth,
                canvasTop + s.vHeaderPadding - penWidth,
                self.iconItem.iconWidth() +
                2 * (s.hHiddenCommentPadding + penWidth),
                self.iconItem.iconHeight() +
                2 * (s.vHiddenCommentPadding + penWidth))
            self.iconItem.setPos(
                movedBaseX + s.hHeaderPadding + s.hHiddenCommentPadding,
                canvasTop + s.vHeaderPadding + s.vHiddenCommentPadding)
            self.iconItem.setToolTip(self.text)

            self.scene.addItem(self)
            self.scene.addItem(self.iconItem)
        else:
            self.setRect(
                movedBaseX + s.hHeaderPadding - penWidth,
                canvasTop + s.vHeaderPadding - penWidth,
                self.textRect.width() +
                2 * s.hTextPadding + 2 * penWidth,
                self.textRect.height() +
                2 * s.vTextPadding + 2 * penWidth)
            self.__sideCommentPath = getNoCellCommentBoxPath(
                movedBaseX + s.hHeaderPadding,
                canvasTop + s.vHeaderPadding,
                self.textRect.width() + 2 * s.hTextPadding,
                self.textRect.height() + 2 * s.vTextPadding,
                s.commentCorner)
            self.scene.addItem(self)

    def __drawDocstring(self):
        """Draws the docstring item"""
        s = self.canvas.settings
        penWidth = s.selectPenWidth - 1
        self.setRect(
            self.baseX - s.scopeRectRadius - penWidth,
            self.baseY - penWidth,
            self.canvas.minWidth - 2 * s.hCellPadding + 2 * penWidth,
            self.height + 2 * penWidth)
        self.scene.addItem(self)
        if s.hidedocstrings:
            self.scene.addItem(self.__docBadge)
            self.__docBadge.moveTo(
                self.baseX - s.scopeRectRadius + penWidth,
                self.baseY + penWidth)

    def draw(self, scene, baseX, baseY):
        """Draws a scope"""
        self.baseX = baseX
        self.baseY = baseY
        self.scene = scene

        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.__drawTopLeft()
        elif self.subKind == ScopeCellElement.DECLARATION:
            self.__drawDeclaration()
        elif self.subKind == ScopeCellElement.COMMENT:
            self.__drawComment()
        elif self.subKind == ScopeCellElement.DOCSTRING:
            self.__drawDocstring()

    def __paintTopLeft(self, painter):
        """Paints the scope rectangle"""
        s = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(), self.borderColor))
        painter.setBrush(QBrush(self.bgColor))
        painter.drawRoundedRect(self.baseX + s.hCellPadding,
                                self.baseY + s.vCellPadding,
                                self.canvas.minWidth - 2 * s.hCellPadding,
                                self.canvas.minHeight - 2 * s.vCellPadding,
                                s.scopeRectRadius, s.scopeRectRadius)

    def __paintDeclaration(self, painter):
        """Paints the scope header"""
        s = self.canvas.settings
        painter.setBrush(QBrush(self.bgColor))
        pen = QPen(self.fgColor)
        painter.setFont(s.monoFont)
        painter.setPen(pen)
        canvasLeft = self.baseX - s.scopeRectRadius
        canvasTop = self.baseY - s.scopeRectRadius
        textHeight = self.textRect.height()
        yShift = 0
        if hasattr(self.ref, 'sideComment'):
            yShift = s.vTextPadding
        painter.drawText(canvasLeft + s.hHeaderPadding,
                         canvasTop + s.vHeaderPadding + yShift,
                         self.textRect.width(), textHeight,
                         Qt.AlignLeft, self.text)

        pen = QPen(self.borderColor)
        pen.setWidth(s.boxLineWidth)
        painter.setPen(pen)

        # If the scope is selected then the line may need to be shorter
        # to avoid covering the outline
        row = self.addr[1] - 1
        column = self.addr[0] - 1
        correction = 0.0
        if self.canvas.cells[row][column].isSelected():
            correction = s.selectPenWidth - 1
        painter.drawLine(canvasLeft + correction, self.baseY + self.height,
                         canvasLeft + self.canvas.minWidth -
                         2 * s.hCellPadding - correction,
                         self.baseY + self.height)

    def __paintComment(self, painter):
        """Paints the side comment"""
        s = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(),
                                          s.commentBorderColor))
        painter.setBrush(QBrush(s.commentBGColor))

        canvasTop = self.baseY - s.scopeRectRadius
        # s.vHeaderPadding below is used intentionally: to have the same
        # spacing on top, bottom and right for the comment box
        movedBaseX = self.canvas.baseX + self.canvas.minWidth - \
            self.width - s.scopeRectRadius - s.vHeaderPadding

        if s.hidecomments:
            painter.drawRoundedRect(
                movedBaseX + s.hHeaderPadding,
                canvasTop + s.vHeaderPadding,
                self.iconItem.iconWidth() + 2 * s.hHiddenCommentPadding,
                self.iconItem.iconHeight() + 2 * s.vHiddenCommentPadding,
                s.hiddenCommentRectRadius, s.hiddenCommentRectRadius)
        else:
            painter.drawPath(self.__sideCommentPath)

            pen = QPen(s.commentFGColor)
            painter.setFont(s.monoFont)
            painter.setPen(pen)
            painter.drawText(
                movedBaseX + s.hHeaderPadding + s.hTextPadding,
                canvasTop + s.vHeaderPadding + s.vTextPadding,
                self.textRect.width(),
                self.textRect.height(),
                Qt.AlignLeft, self.text)

    def __paintDocstring(self, painter):
        """Paints the docstring"""
        s = self.canvas.settings
        painter.setBrush(QBrush(self.bgColor))

        canvasLeft = self.baseX - s.scopeRectRadius

        if self.isSelected():
            selectPen = QPen(s.selectColor)
            selectPen.setWidth(s.selectPenWidth)
            selectPen.setJoinStyle(Qt.RoundJoin)
            painter.setPen(selectPen)
            painter.drawRect(canvasLeft, self.baseY,
                             self.canvas.minWidth - 2 * s.hCellPadding,
                             self.height)
        else:
            # If the scope is selected then the line may need to be shorter
            # to avoid covering the outline
            row = self.addr[1] - 2
            column = self.addr[0] - 1
            correction = 0.0
            if self.canvas.cells[row][column].isSelected():
                correction = s.selectPenWidth - 1

            # The background could also be custom
            pen = QPen(self.bgColor)
            pen.setWidth(s.boxLineWidth)
            pen.setJoinStyle(Qt.MiterJoin)
            painter.setPen(pen)

            dsCorr = float(s.boxLineWidth)
            if self.canvas.cells[row][column].isSelected():
                dsCorr = float(s.selectPenWidth) / 2.0 + \
                    float(s.boxLineWidth) / 2.0
            painter.drawRect(float(canvasLeft) + dsCorr,
                             self.baseY + s.boxLineWidth,
                             float(self.canvas.minWidth) -
                             2.0 * float(s.hCellPadding) - 2.0 * dsCorr,
                             self.height - 2 * s.boxLineWidth)

            pen = QPen(self.borderColor)
            pen.setWidth(s.boxLineWidth)
            painter.setPen(pen)
            painter.drawLine(canvasLeft + correction,
                             self.baseY + self.height,
                             canvasLeft + self.canvas.minWidth -
                             2 * s.hCellPadding - correction,
                             self.baseY + self.height)

        if not s.hidedocstrings:
            pen = QPen(self.fgColor)
            painter.setFont(s.monoFont)
            painter.setPen(pen)
            painter.drawText(canvasLeft + s.hHeaderPadding,
                             self.baseY + s.vHeaderPadding,
                             self.canvas.width - 2 * s.hHeaderPadding,
                             self.height - 2 * s.vHeaderPadding,
                             Qt.AlignLeft, self.text)

    def paint(self, painter, option, widget):
        """Draws the corresponding scope element"""
        del option
        del widget

        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.__paintTopLeft(painter)
        elif self.subKind == ScopeCellElement.DECLARATION:
            self.__paintDeclaration(painter)
        elif self.subKind == ScopeCellElement.COMMENT:
            self.__paintComment(painter)
        elif self.subKind == ScopeCellElement.DOCSTRING:
            self.__paintDocstring(painter)

    def hoverEnterEvent(self, event):
        """Handling mouse enter event"""
        del event
        if self.__navBarUpdate:
            self.__navBarUpdate(self.getCanvasTooltip())

    def hoverLeaveEvent(self, event):
        """Handling mouse enter event"""
        del event
        # if self.__navBarUpdate:
        #     self.__navBarUpdate("")

    def __str__(self):
        """Debugging support"""
        return CellElement.__str__(self) + \
               "(" + scopeCellElementToString(self.subKind) + ")"

    def isComment(self):
        """True if it is a comment"""
        return self.subKind == self.COMMENT

    def isDocstring(self):
        """True if it is a docstring"""
        return self.subKind == self.DOCSTRING

    def mouseDoubleClickEvent(self, event):
        """Jump to the appropriate line in the text editor"""
        if self.subKind == self.COMMENT:
            CellElement.mouseDoubleClickEvent(
                self, event, pos=self.ref.sideComment.beginPos)
        elif self.subKind == self.DOCSTRING:
            CellElement.mouseDoubleClickEvent(
                self, event, pos=self.ref.docstring.body.beginPos)
        elif self.subKind == self.DECLARATION:
            if self.kind == CellElement.FILE_SCOPE:
                CellElement.mouseDoubleClickEvent(
                    self, event, line=1, pos=1)
            else:
                CellElement.mouseDoubleClickEvent(
                    self, event, line=self.ref.body.beginLine,
                    pos=self.ref.body.beginPos)

    def getTopLeftItem(self):
        """Provides a top left corner item"""
        if self.subKind == self.DECLARATION:
            # The scope is at (x-1, y-1) location
            column = self.addr[0] - 1
            row = self.addr[1] - 1
            return self.canvas.cells[row][column]
        raise Exception("Logical error: the getTopLeftItem() is designed "
                        "to be called for the DECLARATION only")

    def getDistance(self, absPos):
        """Provides a distance between the absPos and the item"""
        if self.subKind == self.DOCSTRING:
            return distance(absPos, self.ref.docstring.begin,
                            self.ref.docstring.end)
        if self.subKind == self.COMMENT:
            return distance(absPos, self.ref.sideComment.begin,
                            self.ref.sideComment.end)
        if self.subKind == self.DECLARATION:
            if self.kind == CellElement.FILE_SCOPE:
                dist = maxsize
                if self.ref.encodingLine:
                    dist = min(dist, distance(absPos,
                                              self.ref.encodingLine.begin,
                                              self.ref.encodingLine.end))
                if self.ref.bangLine:
                    dist = min(dist, distance(absPos,
                                              self.ref.bangLine.begin,
                                              self.ref.bangLine.end))
                return dist
            # Not a file scope
            return distance(absPos, self.ref.body.begin,
                            self.ref.body.end)
        return maxsize

    def getLineDistance(self, line):
        """Provides a distance between the line and the item"""
        if self.subKind == self.DOCSTRING:
            return distance(line, self.ref.docstring.beginLine,
                            self.ref.docstring.endLine)
        if self.subKind == self.COMMENT:
            return distance(line, self.ref.sideComment.beginLine,
                            self.ref.sideComment.endLine)
        if self.subKind == self.DECLARATION:
            if self.kind == CellElement.FILE_SCOPE:
                dist = maxsize
                if self.ref.encodingLine:
                    dist = min(dist, distance(line,
                                              self.ref.encodingLine.beginLine,
                                              self.ref.encodingLine.endLine))
                if self.ref.bangLine:
                    dist = min(dist, distance(line,
                                              self.ref.bangLine.beginLine,
                                              self.ref.bangLine.endLine))
                return dist
            # Not a file scope
            return distance(line, self.ref.body.beginLine,
                            self.ref.body.endLine)
        return maxsize

    def getFirstLine(self):
        """Provides the first line"""
        if self.isDocstring():
            line = maxsize
            if self.ref.docstring.leadingCMLComments:
                line = CMLVersion.getFirstLine(self.ref.leadingCMLComments)
            if self.ref.docstring.leadingComment:
                if self.ref.docstring.leadingComment.parts:
                    line = min(
                        self.ref.docstring.leadingComment.parts[0].beginLine,
                        line)
            return min(self.ref.docstring.beginLine, line)
        return CellElement.getFirstLine(self)

    def getTooltipSuffix(self):
        """Provides the selection tooltip suffix"""
        if self.subKind == self.TOP_LEFT:
            return ' at ' + CellElement.getLinesSuffix(self.getLineRange())
        if self.subKind == self.DOCSTRING:
            return ' docstring at ' + \
                CellElement.getLinesSuffix(self.getLineRange())
        if self.subKind == self.COMMENT:
            return ' side comment at ' + \
                CellElement.getLinesSuffix(self.getLineRange())
        return ' scope (' + scopeCellElementToString(self.subKind) + ')'


_scopeCellElementToString = {
    ScopeCellElement.TOP_LEFT: "TOP_LEFT",
    ScopeCellElement.DECLARATION: "DECLARATION",
    ScopeCellElement.COMMENT: "COMMENT",
    ScopeCellElement.DOCSTRING: "DOCSTRING"}


def scopeCellElementToString(kind):
    """Provides a string representation of a element kind"""
    return _scopeCellElementToString[kind]


class FileScopeCell(ScopeCellElement):

    """Represents a file scope element"""

    def __init__(self, ref, canvas, x, y, subKind):
        ScopeCellElement.__init__(self, ref, canvas, x, y, subKind,
                                  canvas.settings.fileScopeBGColor,
                                  canvas.settings.fileScopeFGColor,
                                  canvas.settings.fileScopeBorderColor)
        self.kind = CellElement.FILE_SCOPE

    def render(self):
        """Renders the file scope element"""
        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.badgeItem = BadgeItem(self, "module")
        return self.renderCell()

    def getLineRange(self):
        """Provides the line range"""
        if self.subKind == self.TOP_LEFT:
            if self.ref.body is None:
                # Special case: the buffer is empty so no body exists
                return [0, 0]
        if self.subKind == self.DOCSTRING:
            return self.ref.docstring.body.getLineRange()
        return CellElement.getLineRange(self)

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.subKind == self.TOP_LEFT:
            if self.ref.body is None:
                # Special case: the buffer is empty so no body exists
                return [0, 0]
        if self.subKind == self.DOCSTRING:
            return self.ref.docstring.body.getAbsPosRange()
        return CellElement.getAbsPosRange(self)

    def getSelectTooltip(self):
        """Provides a file scope selected tooltip"""
        return 'Module' + self.getTooltipSuffix()


class FunctionScopeCell(ScopeCellElement):

    """Represents a function scope element"""

    def __init__(self, ref, canvas, x, y, subKind):
        ScopeCellElement.__init__(self, ref, canvas, x, y, subKind,
                                  canvas.settings.funcScopeBGColor,
                                  canvas.settings.funcScopeFGColor,
                                  canvas.settings.funcScopeBorderColor)
        self.kind = CellElement.FUNC_SCOPE

    def getSideComment(self):
        """Provides a side comment"""
        # The comment may start not at the first line of the function
        if self.canvas.settings.hidecomments:
            return self.ref.sideComment.getDisplayValue()

        linesBefore = self.ref.sideComment.beginLine - \
                      self.ref.name.beginLine
        sideComment = '\n' * linesBefore + \
                      self.ref.sideComment.getDisplayValue()

        # The comment may stop before the end of the arguments list
        linesAfter = self.ref.arguments.endLine - self.ref.sideComment.endLine
        return sideComment + '\n' * linesAfter

    def render(self):
        """Renders the function scope element"""
        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.badgeItem = BadgeItem(self, "def")
        return self.renderCell()

    def getLineRange(self):
        """Provides the line range"""
        if self.subKind == self.DOCSTRING:
            return self.ref.docstring.body.getLineRange()
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getLineRange()
        return CellElement.getLineRange(self)

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.subKind == self.DOCSTRING:
            return self.ref.docstring.body.getAbsPosRange()
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getAbsPosRange()
        return CellElement.getAbsPosRange(self)

    def getSelectTooltip(self):
        """Provides a selected function block tooltip"""
        return 'Function' + self.getTooltipSuffix()


class ClassScopeCell(ScopeCellElement):

    """Represents a class scope element"""

    def __init__(self, ref, canvas, x, y, subKind):
        ScopeCellElement.__init__(self, ref, canvas, x, y, subKind,
                                  canvas.settings.classScopeBGColor,
                                  canvas.settings.classScopeFGColor,
                                  canvas.settings.classScopeBorderColor)
        self.kind = CellElement.CLASS_SCOPE

    def getSideComment(self):
        """Provides a side comment"""
        # The comment may start not at the first line of the class
        if self.canvas.settings.hidecomments:
            return self.ref.sideComment.getDisplayValue()

        linesBefore = self.ref.sideComment.beginLine - \
                      self.ref.name.beginLine
        sideComment = '\n' * linesBefore + \
                      self.ref.sideComment.getDisplayValue()
        if self.ref.baseClasses is None:
            lastLine = self.ref.name.endLine
        else:
            lastLine = self.ref.baseClasses.endLine

        # The comment may stop before the end of the arguments list
        linesAfter = lastLine - self.ref.sideComment.endLine
        return sideComment + '\n' * linesAfter

    def render(self):
        """Renders the class scope element"""
        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.badgeItem = BadgeItem(self, "class")
        return self.renderCell()

    def getLineRange(self):
        """Provides the line range"""
        if self.subKind == self.DOCSTRING:
            return self.ref.docstring.body.getLineRange()
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getLineRange()
        return CellElement.getLineRange(self)

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.subKind == self.DOCSTRING:
            return self.ref.docstring.body.getAbsPosRange()
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getAbsPosRange()
        return CellElement.getAbsPosRange(self)

    def getSelectTooltip(self):
        """Provides the selection tooltip"""
        return 'Class' + self.getTooltipSuffix()


class ForScopeCell(ScopeCellElement):

    """Represents a for-loop scope element"""

    def __init__(self, ref, canvas, x, y, subKind):
        ScopeCellElement.__init__(self, ref, canvas, x, y, subKind,
                                  canvas.settings.forScopeBGColor,
                                  canvas.settings.forScopeFGColor,
                                  canvas.settings.forScopeBorderColor)
        self.kind = CellElement.FOR_SCOPE

    def getSideComment(self):
        """Provides a side comment"""
        # The comment may start not at the first line of the function
        if self.canvas.settings.hidecomments:
            return self.ref.sideComment.getDisplayValue()

        linesBefore = self.ref.sideComment.beginLine - \
                      self.ref.iteration.beginLine
        sideComment = '\n' * linesBefore + \
                      self.ref.sideComment.getDisplayValue()

        # The comment may stop before the end of the arguments list
        linesAfter = self.ref.iteration.endLine - self.ref.sideComment.endLine
        return sideComment + '\n' * linesAfter

    def render(self):
        """Renders the for-loop scope element"""
        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.badgeItem = BadgeItem(self, "for")
        return self.renderCell()

    def getLineRange(self):
        """Provides the line range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getLineRange()
        return CellElement.getLineRange(self)

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getAbsPosRange()
        return CellElement.getAbsPosRange(self)

    def getSelectTooltip(self):
        """Provides a selected for block tooltip"""
        return 'For loop' + self.getTooltipSuffix()


class WhileScopeCell(ScopeCellElement):

    """Represents a while-loop scope element"""

    def __init__(self, ref, canvas, x, y, subKind):
        ScopeCellElement.__init__(self, ref, canvas, x, y, subKind,
                                  canvas.settings.whileScopeBGColor,
                                  canvas.settings.whileScopeFGColor,
                                  canvas.settings.whileScopeBorderColor)
        self.kind = CellElement.WHILE_SCOPE

    def getSideComment(self):
        """Provides a side comment"""
        # The comment may start not at the first line of the function
        if self.canvas.settings.hidecomments:
            return self.ref.sideComment.getDisplayValue()

        linesBefore = self.ref.sideComment.beginLine - \
                      self.ref.condition.beginLine
        sideComment = '\n' * linesBefore + \
                      self.ref.sideComment.getDisplayValue()

        # The comment may stop before the end of the arguments list
        linesAfter = self.ref.condition.endLine - self.ref.sideComment.endLine
        return sideComment + '\n' * linesAfter

    def render(self):
        """Renders the while-loop scope element"""
        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.badgeItem = BadgeItem(self, "while")
        return self.renderCell()

    def getLineRange(self):
        """Provides the lineRange"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getLineRange()
        return CellElement.getLineRange(self)

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getAbsPosRange()
        return CellElement.getAbsPosRange(self)

    def getSelectTooltip(self):
        """Provides the selected while scope element tooltip"""
        return 'While loop' + self.getTooltipSuffix()


class TryScopeCell(ScopeCellElement):

    """Represents a try-except scope element"""

    def __init__(self, ref, canvas, x, y, subKind):
        ScopeCellElement.__init__(self, ref, canvas, x, y, subKind,
                                  canvas.settings.tryScopeBGColor,
                                  canvas.settings.tryScopeFGColor,
                                  canvas.settings.tryScopeBorderColor)
        self.kind = CellElement.TRY_SCOPE

    def getSideComment(self):
        """Provides a side comment"""
        return self.ref.sideComment.getDisplayValue()

    def render(self):
        """Renders the try scope element"""
        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.badgeItem = BadgeItem(self, "try")
        return self.renderCell()

    def getLineRange(self):
        """Provides the line range"""
        if self.subKind == self.TOP_LEFT:
            beginLine = self.ref.body.beginLine
            _, endLine = self.ref.suite[-1].getLineRange()
            return [beginLine, endLine]
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getLineRange()
        return CellElement.getLineRange(self)

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.subKind == self.TOP_LEFT:
            begin = self.ref.body.begin
            _, end = self.ref.suite[-1].getAbsPosRange()
            return [begin, end]
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getAbsPosRange()
        return CellElement.getAbsPosRange(self)

    def getSelectTooltip(self):
        """Provides the selected try block tooltip"""
        return 'Try' + self.getTooltipSuffix()


class WithScopeCell(ScopeCellElement):

    """Represents a with scope element"""

    def __init__(self, ref, canvas, x, y, subKind):
        ScopeCellElement.__init__(self, ref, canvas, x, y, subKind,
                                  canvas.settings.withScopeBGColor,
                                  canvas.settings.withScopeFGColor,
                                  canvas.settings.withScopeBorderColor)
        self.kind = CellElement.WITH_SCOPE

    def getSideComment(self):
        """Provides a side comment"""
        # The comment may start not at the first line of the function
        if self.canvas.settings.hidecomments:
            return self.ref.sideComment.getDisplayValue()

        linesBefore = self.ref.sideComment.beginLine - \
                      self.ref.items.beginLine
        sideComment = '\n' * linesBefore + \
                      self.ref.sideComment.getDisplayValue()

        # The comment may stop before the end of the arguments list
        linesAfter = self.ref.items.endLine - self.ref.sideComment.endLine
        return sideComment + '\n' * linesAfter

    def render(self):
        """Renders the with block"""
        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.badgeItem = BadgeItem(self, "with")
        return self.renderCell()

    def getLineRange(self):
        """Provides the line range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getLineRange()
        return CellElement.getLineRange(self)

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getAbsPosRange()
        return CellElement.getAbsPosRange(self)

    def getSelectTooltip(self):
        """Provides the selected with block tooltip"""
        return 'With' + self.getTooltipSuffix()


class DecoratorScopeCell(ScopeCellElement):

    """Represents a decorator scope element"""

    def __init__(self, ref, canvas, x, y, subKind):
        ScopeCellElement.__init__(self, ref, canvas, x, y, subKind,
                                  canvas.settings.decorScopeBGColor,
                                  canvas.settings.decorScopeFGColor,
                                  canvas.settings.decorScopeBorderColor)
        self.kind = CellElement.DECOR_SCOPE

    def getSideComment(self):
        """Provides a side comment"""
        # The comment may start not at the first line of the function
        if self.canvas.settings.hidecomments:
            return self.ref.sideComment.getDisplayValue()

        linesBefore = self.ref.sideComment.beginLine - \
                      self.ref.name.beginLine
        sideComment = '\n' * linesBefore + \
                      self.ref.sideComment.getDisplayValue()
        if self.ref.arguments is None:
            lastLine = self.ref.name.endLine
        else:
            lastLine = self.ref.arguments.endLine

        # The comment may stop before the end of the arguments list
        linesAfter = lastLine - self.ref.sideComment.endLine
        return sideComment + '\n' * linesAfter

    def render(self):
        """Renders the decorator"""
        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.badgeItem = BadgeItem(self, " @ ")
        return self.renderCell()

    def getLineRange(self):
        """Provides the line range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getLineRange()
        return CellElement.getLineRange(self)

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getAbsPosRange()
        return CellElement.getAbsPosRange(self)

    def getSelectTooltip(self):
        """Provides the selected decorator tooltip"""
        return 'Decorator' + self.getTooltipSuffix()


class ElseScopeCell(ScopeCellElement):

    """Represents an else scope element"""

    FOR_STATEMENT = 0
    WHILE_STATEMENT = 1
    TRY_STATEMENT = 2

    def __init__(self, ref, canvas, x, y, subKind, statement):
        if statement == self.FOR_STATEMENT:
            bgColor = canvas.settings.forElseScopeBGColor
            fgColor = canvas.settings.forElseScopeFGColor
            borderColor = canvas.settings.forElseScopeBorderColor
        elif statement == self.WHILE_STATEMENT:
            bgColor = canvas.settings.whileElseScopeBGColor
            fgColor = canvas.settings.whileElseScopeFGColor
            borderColor = canvas.settings.whileElseScopeBorderColor
        else:
            # TRY_STATEMENT
            bgColor = canvas.settings.tryElseScopeBGColor
            fgColor = canvas.settings.tryElseScopeFGColor
            borderColor = canvas.settings.tryElseScopeBorderColor
        ScopeCellElement.__init__(self, ref, canvas, x, y, subKind,
                                  bgColor, fgColor, borderColor)
        self.kind = CellElement.ELSE_SCOPE
        self.statement = statement
        self.after = self

    def getSideComment(self):
        """Provides a side comment"""
        return self.ref.sideComment.getDisplayValue()

    def render(self):
        """Renders the else block"""
        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.badgeItem = BadgeItem(self, "else")
        return self.renderCell()

    def getLineRange(self):
        """Provides the line range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getLineRange()
        return CellElement.getLineRange(self)

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getAbsPosRange()
        return CellElement.getAbsPosRange(self)

    def getSelectTooltip(self):
        """Provides the selection tooltip"""
        return 'Else' + self.getTooltipSuffix()


class ForElseScopeCell(ElseScopeCell):

    """Else scope which is bound to a for loop"""

    def __init__(self, ref, canvas, x, y, subKind):
        ElseScopeCell.__init__(self, ref, canvas, x, y, subKind,
                               ElseScopeCell.FOR_STATEMENT)


class WhileElseScopeCell(ElseScopeCell):

    """Else scope which is bound to a while loop"""

    def __init__(self, ref, canvas, x, y, subKind):
        ElseScopeCell.__init__(self, ref, canvas, x, y, subKind,
                               ElseScopeCell.WHILE_STATEMENT)


class TryElseScopeCell(ElseScopeCell):

    """Else scope which is bound to a try block"""

    def __init__(self, ref, canvas, x, y, subKind):
        ElseScopeCell.__init__(self, ref, canvas, x, y, subKind,
                               ElseScopeCell.TRY_STATEMENT)


class ExceptScopeCell(ScopeCellElement):

    """Represents an except scope element"""

    def __init__(self, ref, canvas, x, y, subKind):
        ScopeCellElement.__init__(self, ref, canvas, x, y, subKind,
                                  canvas.settings.exceptScopeBGColor,
                                  canvas.settings.exceptScopeFGColor,
                                  canvas.settings.exceptScopeBorderColor)
        self.kind = CellElement.EXCEPT_SCOPE

    def getSideComment(self):
        """Provides a side comment"""
        # The comment may start not at the first line of the except
        if self.ref.clause is None or self.canvas.settings.hidecomments:
            return self.ref.sideComment.getDisplayValue()

        linesBefore = self.ref.sideComment.beginLine - \
                      self.ref.clause.beginLine
        sideComment = '\n' * linesBefore + \
                      self.ref.sideComment.getDisplayValue()
        lastLine = self.ref.clause.endLine

        # The comment may stop before the end of the arguments list
        linesAfter = lastLine - self.ref.sideComment.endLine
        return sideComment + '\n' * linesAfter

    def render(self):
        """Renders the except block"""
        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.badgeItem = BadgeItem(self, "except")
        return self.renderCell()

    def getLineRange(self):
        """Provides the line range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getLineRange()
        return CellElement.getLineRange(self)

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getAbsPosRange()
        return CellElement.getAbsPosRange(self)

    def getSelectTooltip(self):
        """Provides the selection tooltip"""
        return 'Except' + self.getTooltipSuffix()


class FinallyScopeCell(ScopeCellElement):

    """Represents a finally scope element"""

    def __init__(self, ref, canvas, x, y, subKind):
        ScopeCellElement.__init__(self, ref, canvas, x, y, subKind,
                                  canvas.settings.finallyScopeBGColor,
                                  canvas.settings.finallyScopeFGColor,
                                  canvas.settings.finallyScopeBorderColor)
        self.kind = CellElement.FINALLY_SCOPE

    def getSideComment(self):
        """Provides the side comment"""
        return self.ref.sideComment.getDisplayValue()

    def render(self):
        """Renders the finally block"""
        if self.subKind == ScopeCellElement.TOP_LEFT:
            self.badgeItem = BadgeItem(self, "finally")
        return self.renderCell()

    def getLineRange(self):
        """Provides the line range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getLineRange()
        return CellElement.getLineRange(self)

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        if self.subKind == self.COMMENT:
            return self.ref.sideComment.getAbsPosRange()
        return CellElement.getAbsPosRange(self)

    def getSelectTooltip(self):
        """Provides a tooltip for the selected finally block"""
        return 'Finally' + self.getTooltipSuffix()




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


"""codimension brief module info cache"""

from os.path import realpath, getmtime, exists
from cdmpyparser import getBriefModuleInfoFromFile


class BriefModuleInfoCache():

    """Provides the module info cache"""

    def __init__(self):
        # abs file path -> (modification time, mod info)
        self.__cache = {}

    def get(self, path):
        """Provides the required modinfo"""
        path = realpath(path)
        try:
            modInfo = self.__cache[path]
            if not exists(path):
                del self.__cache[path]
                raise Exception("Cannot open " + path)

            lastModTime = getmtime(path)
            if lastModTime <= modInfo[0]:
                return modInfo[1]

            # update the key
            info = getBriefModuleInfoFromFile(path)
            self.__cache[path] = (lastModTime, info)
            return info
        except KeyError:
            if not exists(path):
                raise Exception("Cannot open " + path)

            info = getBriefModuleInfoFromFile(path)
            self.__cache[path] = (getmtime(path), info)
            return info

    def remove(self, path):
        """Removes one file from the map"""
        path = realpath(path)
        try:
            del self.__cache[path]
        except KeyError:
            return

    def clear(self):
        """Clears the cache"""
        self.__cache = {}



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2019  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Utils to work with binary files"""

import logging
from .globals import GlobalData


def getHexdump(fileName):
    """Provides the hex dumped file content or None"""
    if not GlobalData().hexdumpAvailable:
        logging.error('hexdump is not available')
        return None

    process = QProcess()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Running/debugging session parameters"""

from copy import deepcopy


# These constants are used throughout the code to identify the type of run
RUN = 0
PROFILE = 1
DEBUG = 2


class RunParameters:

    """Stores the script run parameters"""

    InheritParentEnv = 0
    InheritParentEnvPlus = 1
    SpecificEnvironment = 2

    def __init__(self):
        self.__params = deepcopy(DEFAULT_RUN_PARAMETERS)

    def __getitem__(self, key):
        return self.__params[key]

    def __setitem__(self, key, value):
        self.__params[key] = value

    def isDefault(self):
        """Returns True if all the values are default"""
        return self.__params == DEFAULT_RUN_PARAMETERS

    def runParamsToJSON(self):
        """Converts the instance to a serializable structure"""
        return {'__class__': 'RunParameters',
                '__values__': self.__params}

    def runParamsFromJSON(self, jsonObj):
        """Populates the values from the json object"""
        self.__params = jsonObj['__values__']


# Default parameters
DEFAULT_RUN_PARAMETERS = {
    # Cmd line arguments
    'arguments': '',

    # Working dir part
    'useScriptLocation': True,
    'specificDir': '',

    # Environment
    'envType': RunParameters.InheritParentEnv,
    'additionToParentEnv': {},
    'specificEnv': {},

    # Path to python
    'useInherited': True,
    'customInterpreter': '',

    # Way to run
    'redirected': True,
    'customTerminal': ''}


# JSON serialization/deserialization support
# implementation idea is taken from here:
# http://www.diveintopython3.net/serializing.html
def runParamsToJSON(pythonObj):
    """Custom serialization"""
    if isinstance(pythonObj, RunParameters):
        return pythonObj.runParamsToJSON()
    raise TypeError(repr(pythonObj) + ' is not JSON serializable')


def runParamsFromJSON(jsonObj):
    """Custom deserialization"""
    if '__class__' in jsonObj:
        if jsonObj['__class__'] == 'RunParameters':
            params = RunParameters()
            params.runParamsFromJSON(jsonObj)
            return params
    return jsonObj



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Codimension main window menu"""


import os.path
from utils.pixmapcache import getIcon
from utils.skin import getSkinsList
from utils.globals import GlobalData
from utils.misc import getIDETemplateFile, getProjectTemplateFile
from utils.settings import CLEAR_AND_REUSE, NO_CLEAR_AND_REUSE, NO_REUSE
from utils.diskvaluesrelay import getRecentFiles
from .qt import QApplication, QMenu, QStyleFactory, QActionGroup, QFontDialog
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase


def getAccelerator(count):
    """Provides an accelerator text for a menu item"""
    if count < 10:
        return "&" + str(count) + ".  "
    return "&" + chr(count - 10 + ord('a')) + ".  "


class MainWindowMenuMixin:

    """Main window menu mixin"""

    def __init__(self):
        pass

    def _initMainMenu(self):
        """Initializes the main menu bar"""
        menuBar = self.menuBar()
        for member in [self.__buildProjectMenu, self.__buildTabMenu,
                       self.__buildEditMenu, self.__buildSearchMenu,
                       self.__buildRunMenu, self.__buildDebugMenu,
                       self.__buildToolsMenu, self.__buildDiagramsMenu,
                       self.__buildViewMenu, self.__buildOptionsMenu,
                       self.__buildPluginsMenu, self.__buildHelpMenu]:
            menuBar.addMenu(member(menuBar))

    def __buildProjectMenu(self, menuBar):
        """Builds a project menu"""
        prjMenu = QMenu("&Project", menuBar)
        prjMenu.setObjectName('project')
        prjMenu.aboutToShow.connect(self.__prjAboutToShow)
        prjMenu.aboutToHide.connect(self.__prjAboutToHide)
        self.__newProjectAct = prjMenu.addAction(
            getIcon('createproject.png'), "&New project",
            self._createNewProject, 'Ctrl+Shift+N')
        self.__openProjectAct = prjMenu.addAction(
            getIcon('project.png'), '&Open project',
            self._openProject, 'Ctrl+Shift+O')
        self._unloadProjectAct = prjMenu.addAction(
            getIcon('unloadproject.png'), '&Unload project',
            self.projectViewer.unloadProject)
        self._projectPropsAct = prjMenu.addAction(
            getIcon('smalli.png'), '&Properties',
            self.projectViewer.projectProperties)
        prjMenu.addSeparator()
        self._prjTemplateMenu = QMenu("Project-specific &template", self)
        self.__createPrjTemplateAct = self._prjTemplateMenu.addAction(
            getIcon('generate.png'), '&Create')
        self.__createPrjTemplateAct.triggered.connect(
            self._onCreatePrjTemplate)
        self.__editPrjTemplateAct = self._prjTemplateMenu.addAction(
            getIcon('edit.png'), '&Edit')
        self.__editPrjTemplateAct.triggered.connect(self._onEditPrjTemplate)
        self._prjTemplateMenu.addSeparator()
        self.__delPrjTemplateAct = self._prjTemplateMenu.addAction(
            getIcon('trash.png'), '&Delete')
        self.__delPrjTemplateAct.triggered.connect(self._onDelPrjTemplate)
        prjMenu.addMenu(self._prjTemplateMenu)
        prjMenu.addSeparator()
        self.__recentPrjMenu = QMenu("&Recent projects", self)
        self.__recentPrjMenu.triggered.connect(self._onRecentPrj)
        prjMenu.addMenu(self.__recentPrjMenu)
        prjMenu.addSeparator()
        self.__quitAct = prjMenu.addAction(
            getIcon('exitmenu.png'), "E&xit codimension",
            QApplication.closeAllWindows, "Ctrl+Q")
        return prjMenu

    def __buildTabMenu(self, menuBar):
        """Build the tab menu"""
        tabMenu = QMenu("&Tab", menuBar)
        tabMenu.setObjectName('tab')
        tabMenu.aboutToShow.connect(self.__tabAboutToShow)
        tabMenu.aboutToHide.connect(self.__tabAboutToHide)
        self.__newTabAct = tabMenu.addAction(
            getIcon('filemenu.png'), "&New tab",
            self.em.newTabClicked, 'Ctrl+N')
        self.__openFileAct = tabMenu.addAction(
            getIcon('filemenu.png'), '&Open file', self._openFile, 'Ctrl+O')
        self.__cloneTabAct = tabMenu.addAction(
            getIcon('clonetabmenu.png'), '&Clone tab', self.em.onClone)
        self.__closeOtherTabsAct = tabMenu.addAction(
            getIcon(''), 'Close oth&er tabs', self.em.onCloseOther)
        self.__closeTabAct = tabMenu.addAction(
            getIcon('closetabmenu.png'), 'Close &tab', self.em.onCloseTab)
        tabMenu.addSeparator()
        self.__saveFileAct = tabMenu.addAction(
            getIcon('savemenu.png'), '&Save', self.em.onSave, 'Ctrl+S')
        self.__saveFileAsAct = tabMenu.addAction(
            getIcon('saveasmenu.png'),
            'Save &as...', self.em.onSaveAs, "Ctrl+Shift+S")
        self.__tabJumpToDefAct = tabMenu.addAction(
            getIcon('definition.png'), "&Jump to definition",
            self._onTabJumpToDef)
        self.__calltipAct = tabMenu.addAction(
            getIcon('calltip.png'), 'Show &calltip', self._onShowCalltip)
        self.__tabJumpToScopeBeginAct = tabMenu.addAction(
            getIcon('jumpupscopemenu.png'),
            'Jump to scope &begin', self._onTabJumpToScopeBegin)
        self.__tabOpenImportAct = tabMenu.addAction(
            getIcon('imports.png'), 'Open &import(s)', self._onTabOpenImport)
        self.__openAsFileAct = tabMenu.addAction(
            getIcon('filemenu.png'), 'O&pen as file', self._onOpenAsFile)
        self.__downloadAndShowAct = tabMenu.addAction(
            getIcon('filemenu.png'), 'Download and show',
            self._onDownloadAndShow)
        self.__openInBrowserAct = tabMenu.addAction(
            getIcon('homepagemenu.png'), 'Open in browser',
            self._onOpenInBrowser)
        tabMenu.addSeparator()
        self.__highlightInPrjAct = tabMenu.addAction(
            getIcon('highlightmenu.png'), 'Highlight in project browser',
            self.em.onHighlightInPrj)
        self.__highlightInFSAct = tabMenu.addAction(
            getIcon('highlightmenu.png'), 'Highlight in file system browser',
            self.em.onHighlightInFS)
        self.__highlightInOutlineAct = tabMenu.addAction(
            getIcon('highlightmenu.png'), 'Highlight in outline browser',
            self._onHighlightInOutline)
        tabMenu.addSeparator()
        self.__recentFilesMenu = QMenu("&Recent files", self)
        self.__recentFilesMenu.triggered.connect(self._onRecentFile)
        tabMenu.addMenu(self.__recentFilesMenu)
        return tabMenu

    def __buildEditMenu(self, menuBar):
        """Builds edit menu"""
        editMenu = QMenu("&Edit", menuBar)
        editMenu.setObjectName('edit')
        editMenu.aboutToShow.connect(self.__editAboutToShow)
        editMenu.aboutToHide.connect(self.__editAboutToHide)
        self.__undoAct = editMenu.addAction(
            getIcon('undo.png'), '&Undo', self._onUndo)
        self.__redoAct = editMenu.addAction(
            getIcon('redo.png'), '&Redo', self._onRedo)
        editMenu.addSeparator()
        self.__cutAct = editMenu.addAction(
            getIcon('cutmenu.png'), 'Cu&t', self._onCut)
        self.__copyAct = editMenu.addAction(
            getIcon('copymenu.png'), '&Copy', self.em.onCopy)
        self.__pasteAct = editMenu.addAction(
            getIcon('pastemenu.png'), '&Paste', self._onPaste)
        self.__selectAllAct = editMenu.addAction(
            getIcon('selectallmenu.png'), 'Select &all', self._onSelectAll)
        editMenu.addSeparator()
        self.__commentAct = editMenu.addAction(
            getIcon('commentmenu.png'), 'C&omment/uncomment', self._onComment)
        self.__duplicateAct = editMenu.addAction(
            getIcon('duplicatemenu.png'), '&Duplicate line', self._onDuplicate)
        self.__autocompleteAct = editMenu.addAction(
            getIcon('autocompletemenu.png'), 'Autoco&mplete',
            self._onAutocomplete)
        self.__expandTabsAct = editMenu.addAction(
            getIcon('expandtabs.png'), 'Expand tabs (&4 spaces)',
            self._onExpandTabs)
        self.__trailingSpacesAct = editMenu.addAction(
            getIcon('trailingws.png'), 'Remove trailing &spaces',
            self._onRemoveTrailingSpaces)
        return editMenu

    def __buildSearchMenu(self, menuBar):
        """Build the search menu"""
        searchMenu = QMenu("&Search", menuBar)
        searchMenu.setObjectName('search')
        searchMenu.aboutToShow.connect(self.__searchAboutToShow)
        searchMenu.aboutToHide.connect(self.__searchAboutToHide)
        self.__searchInFilesAct = searchMenu.addAction(
            getIcon('findindir.png'), "Find in file&s",
            self.findInFilesClicked, "Ctrl+Shift+F")
        searchMenu.addSeparator()
        self._findNameMenuAct = searchMenu.addAction(
            getIcon('findname.png'), 'Find &name in project',
            self.findNameClicked, 'Alt+Shift+S')
        self._findProjectFileAct = searchMenu.addAction(
            getIcon('findfile.png'), 'Find &project file',
            self.findFileClicked, 'Alt+Shift+O')
        searchMenu.addSeparator()
        self.__findOccurencesAct = searchMenu.addAction(
            getIcon('findindir.png'), 'Find &occurrences',
            self._onFindOccurences)
        self.__findAct = searchMenu.addAction(
            getIcon('findindir.png'), '&Find...', self._onFind)
        self.__findNextAct = searchMenu.addAction(
            getIcon('1rightarrow.png'), "&Next highlight", self._onFindNext)
        self.__findPrevAct = searchMenu.addAction(
            getIcon('1leftarrow.png'), "Pre&vious highlight",
            self._onFindPrevious)
        self.__replaceAct = searchMenu.addAction(
            getIcon('replace.png'), '&Replace...', self._onReplace)
        self.__goToLineAct = searchMenu.addAction(
            getIcon('gotoline.png'), '&Go to line...', self._onGoToLine)
        return searchMenu

    def __buildRunMenu(self, menuBar):
        """Build the run menu"""
        runMenu = QMenu("&Run", menuBar)
        runMenu.setObjectName('run')
        runMenu.aboutToShow.connect(self.__runAboutToShow)
        self.__prjRunAct = runMenu.addAction(
            getIcon('run.png'), 'Run &project main script',
            self.onRunProject)
        self.__prjRunDlgAct = runMenu.addAction(
            getIcon('detailsdlg.png'), 'Run p&roject main script...',
            self.onRunProjectDlg)
        self._tabRunAct = runMenu.addAction(
            getIcon('run.png'), 'Run &tab script', self.onRunTab)
        self._tabRunDlgAct = runMenu.addAction(
            getIcon('detailsdlg.png'), 'Run t&ab script...', self.onRunTabDlg)
        runMenu.addSeparator()
        self.__prjProfileAct = runMenu.addAction(
            getIcon('profile.png'), 'Profile project main script',
            self.onProfileProject)
        self.__prjProfileDlgAct = runMenu.addAction(
            getIcon('profile.png'), 'Profile project main script...',
            self.onProfileProjectDlg)
        self._tabProfileAct = runMenu.addAction(
            getIcon('profile.png'), 'Profile tab script', self.onProfileTab)
        self._tabProfileDlgAct = runMenu.addAction(
            getIcon('profile.png'), 'Profile tab script...',
            self.onProfileTabDlg)
        return runMenu

    def __buildDebugMenu(self, menuBar):
        """Build the debug menu"""
        dbgMenu = QMenu("Debu&g", menuBar)
        dbgMenu.setObjectName('debug')
        dbgMenu.aboutToShow.connect(self.__debugAboutToShow)
        self._prjDebugAct = dbgMenu.addAction(
            getIcon('debugger.png'), 'Debug &project main script',
            self.onDebugProject, "Shift+F5")
        self._prjDebugDlgAct = dbgMenu.addAction(
            getIcon('detailsdlg.png'), 'Debug p&roject main script...',
            self.onDebugProjectDlg, "Ctrl+Shift+F5")
        self._tabDebugAct = dbgMenu.addAction(
            getIcon('debugger.png'), 'Debug &tab script',
            self.onDebugTab, "F5")
        self._tabDebugDlgAct = dbgMenu.addAction(
            getIcon('detailsdlg.png'), 'Debug t&ab script...',
            self.onDebugTabDlg, "Ctrl+F5")
        dbgMenu.addSeparator()
        self._debugStopAct = dbgMenu.addAction(
            getIcon('dbgstop.png'), 'Stop debugging session',
            self._onStopDbgSession, "F10")
        self._debugStopAct.setEnabled(False)
        self._debugRestartAct = dbgMenu.addAction(
            getIcon('dbgrestart.png'), 'Restart session',
            self._onRestartDbgSession, "F4")
        self._debugRestartAct.setEnabled(False)
        dbgMenu.addSeparator()
        self._debugContinueAct = dbgMenu.addAction(
            getIcon('dbggo.png'), 'Continue', self._onDbgGo, "F6")
        self._debugContinueAct.setEnabled(False)
        self._debugStepInAct = dbgMenu.addAction(
            getIcon('dbgstepinto.png'), 'Step in', self._onDbgStepInto, "F7")
        self._debugStepInAct.setEnabled(False)
        self._debugStepOverAct = dbgMenu.addAction(
            getIcon('dbgnext.png'), 'Step over', self._onDbgNext, "F8")
        self._debugStepOverAct.setEnabled(False)
        self._debugStepOutAct = dbgMenu.addAction(
            getIcon('dbgreturn.png'), 'Step out', self._onDbgReturn, "F9")
        self._debugStepOutAct.setEnabled(False)
        self._debugRunToCursorAct = dbgMenu.addAction(
            getIcon('dbgruntoline.png'), 'Run to cursor',
            self._onDbgRunToLine, "Shift+F6")
        self._debugRunToCursorAct.setEnabled(False)
        self._debugJumpToCurrentAct = dbgMenu.addAction(
            getIcon('dbgtocurrent.png'), 'Show current line',
            self._onDbgJumpToCurrent, "Ctrl+W")
        self._debugJumpToCurrentAct.setEnabled(False)
        dbgMenu.addSeparator()

        self.__dumpDbgSettingsMenu = QMenu("Dump debug settings", self)
        dbgMenu.addMenu(self.__dumpDbgSettingsMenu)
        self._debugDumpSettingsAct = self.__dumpDbgSettingsMenu.addAction(
            getIcon('dbgsettings.png'), 'Debug session settings',
            self._onDumpDebugSettings)
        self._debugDumpSettingsAct.setEnabled(False)
        self._debugDumpSettingsEnvAct = self.__dumpDbgSettingsMenu.addAction(
            getIcon('detailsdlg.png'),
            'Session settings with complete environment',
            self._onDumpFullDebugSettings)
        self._debugDumpSettingsEnvAct.setEnabled(False)
        self.__dumpDbgSettingsMenu.addSeparator()
        self.__debugDumpScriptSettingsAct = \
            self.__dumpDbgSettingsMenu.addAction(
                getIcon('dbgsettings.png'), 'Current script settings',
                self._onDumpScriptDebugSettings)
        self.__debugDumpScriptSettingsAct.setEnabled(False)
        self.__debugDumpScriptSettingsEnvAct = \
            self.__dumpDbgSettingsMenu.addAction(
                getIcon('detailsdlg.png'),
                'Current script settings with complete environment',
                self._onDumpScriptFullDebugSettings)
        self.__debugDumpScriptSettingsEnvAct.setEnabled(False)
        self.__dumpDbgSettingsMenu.addSeparator()
        self.__debugDumpProjectSettingsAct = \
            self.__dumpDbgSettingsMenu.addAction(
                getIcon('dbgsettings.png'), 'Project main script settings',
                self._onDumpProjectDebugSettings)
        self.__debugDumpProjectSettingsAct.setEnabled(False)
        self.__debugDumpPrjSettingsEnvAct = \
            self.__dumpDbgSettingsMenu.addAction(
                getIcon('detailsdlg.png'),
                'Project script settings with complete environment',
                self._onDumpProjectFullDebugSettings)
        self.__debugDumpPrjSettingsEnvAct.setEnabled(False)
        self.__dumpDbgSettingsMenu.aboutToShow.connect(
            self.__onDumpDbgSettingsAboutToShow)
        return dbgMenu

    def __buildToolsMenu(self, menuBar):
        """Build the tools menu"""
        toolsMenu = QMenu("T&ools", menuBar)
        toolsMenu.setObjectName('tools')
        toolsMenu.aboutToShow.connect(self.__toolsAboutToShow)

        self._deadCodeMenuAct = toolsMenu.addAction(
            getIcon('deadcode.png'), 'Find project &dead code',
            self.projectDeadCodeClicked, 'Alt+Shift+D')
        self._tabDeadCodeAct = toolsMenu.addAction(
            getIcon('deadcode.png'), 'Find tab dead code',
            self.tabDeadCodeClicked, '')
        toolsMenu.addSeparator()
        self.disasmMenu = QMenu('Disassembly', self)
        self.disasmMenu.setIcon(getIcon('disassembly.png'))
        self.disasmAct0 = self.disasmMenu.addAction(
            getIcon(''), 'Disassembly (no optimization)',
            self.onDisasm0)
        self.disasmAct1 = self.disasmMenu.addAction(
            getIcon(''), 'Disassembly (optimization level 1)',
            self.onDisasm1)
        self.disasmAct2 = self.disasmMenu.addAction(
            getIcon(''), 'Disassembly (optimization level 2)',
            self.onDisasm2)
        toolsMenu.addMenu(self.disasmMenu)
        return toolsMenu

    def __buildDiagramsMenu(self, menuBar):
        """Builds the diagram menu"""
        diagramsMenu = QMenu("&Diagrams", menuBar)
        diagramsMenu.setObjectName('diagrams')
        diagramsMenu.aboutToShow.connect(self.__diagramsAboutToShow)
        self._prjImportDgmAct = diagramsMenu.addAction(
            getIcon('importsdiagram.png'), '&Project imports diagram',
            self._onImportDgm)
        self._prjImportsDgmDlgAct = diagramsMenu.addAction(
            getIcon('detailsdlg.png'), 'P&roject imports diagram...',
            self._onImportDgmTuned)
        self.__tabImportDgmAct = diagramsMenu.addAction(
            getIcon('importsdiagram.png'), '&Tab imports diagram',
            self._onTabImportDgm)
        self.__tabImportDgmDlgAct = diagramsMenu.addAction(
            getIcon('detailsdlg.png'), 'T&ab imports diagram...',
            self._onTabImportDgmTuned)
        return diagramsMenu

    def __buildViewMenu(self, menuBar):
        """Build the view menu"""
        viewMenu = QMenu("&View", menuBar)
        viewMenu.setObjectName('view')
        viewMenu.aboutToShow.connect(self.__viewAboutToShow)
        viewMenu.aboutToHide.connect(self.__viewAboutToHide)
        self.__shrinkBarsAct = viewMenu.addAction(
            getIcon('shrinkmenu.png'), "&Hide sidebars",
            self._onMaximizeEditor, 'F11')
        self.__leftSideBarMenu = QMenu("&Left sidebar", self)
        self.__leftSideBarMenu.triggered.connect(self._activateSideTab)
        self.__prjBarAct = self.__leftSideBarMenu.addAction(
            getIcon('project.png'), 'Activate &project tab')
        self.__prjBarAct.setData('project')
        self.__recentBarAct = self.__leftSideBarMenu.addAction(
            getIcon('project.png'), 'Activate &recent tab')
        self.__recentBarAct.setData('recent')
        self.__classesBarAct = self.__leftSideBarMenu.addAction(
            getIcon('class.png'), 'Activate &classes tab')
        self.__classesBarAct.setData('classes')
        self.__funcsBarAct = self.__leftSideBarMenu.addAction(
            getIcon('fx.png'), 'Activate &functions tab')
        self.__funcsBarAct.setData('functions')
        self.__globsBarAct = self.__leftSideBarMenu.addAction(
            getIcon('globalvar.png'), 'Activate &globals tab')
        self.__globsBarAct.setData('globals')
        self.__leftSideBarMenu.addSeparator()
        self.__hideLeftSideBarAct = self.__leftSideBarMenu.addAction(
            getIcon(""), '&Hide left sidebar', self._leftSideBar.shrink)
        viewMenu.addMenu(self.__leftSideBarMenu)

        self.__rightSideBarMenu = QMenu("&Right sidebar", self)
        self.__rightSideBarMenu.triggered.connect(self._activateSideTab)
        self.__outlineBarAct = self.__rightSideBarMenu.addAction(
            getIcon('filepython.png'), 'Activate file &outline tab')
        self.__outlineBarAct.setData('fileoutline')
        self.__debugBarAct = self.__rightSideBarMenu.addAction(
            getIcon(''), 'Activate &debug tab')
        self.__debugBarAct.setData('debugger')
        self.__excptBarAct = self.__rightSideBarMenu.addAction(
            getIcon(''), 'Activate &exceptions tab')
        self.__excptBarAct.setData('excptions')
        self.__bpointBarAct = self.__rightSideBarMenu.addAction(
            getIcon(''), 'Activate &breakpoints tab')
        self.__bpointBarAct.setData('breakpoints')
        self.__calltraceBarAct = self.__rightSideBarMenu.addAction(
            getIcon(''), 'Activate &call trace tab')
        self.__calltraceBarAct.setData('calltrace')
        self.__rightSideBarMenu.addSeparator()
        self.__hideRightSideBarAct = self.__rightSideBarMenu.addAction(
            getIcon(""), '&Hide right sidebar', self._rightSideBar.shrink)
        viewMenu.addMenu(self.__rightSideBarMenu)

        self.__bottomSideBarMenu = QMenu("&Bottom sidebar", self)
        self.__bottomSideBarMenu.triggered.connect(self._activateSideTab)
        self.__logBarAct = self.__bottomSideBarMenu.addAction(
            getIcon('logviewer.png'), 'Activate &log tab')
        self.__logBarAct.setData('log')
        self.__searchBarAct = self.__bottomSideBarMenu.addAction(
            getIcon('findindir.png'), 'Activate &search tab')
        self.__searchBarAct.setData('search')
        self.__bottomSideBarMenu.addSeparator()
        self.__hideBottomSideBarAct = self.__bottomSideBarMenu.addAction(
            getIcon(""), '&Hide bottom sidebar', self._bottomSideBar.shrink)
        viewMenu.addMenu(self.__bottomSideBarMenu)
        viewMenu.addSeparator()
        self.__zoomInAct = viewMenu.addAction(
            getIcon('zoomin.png'), 'Zoom &in', self.em.zoomIn)
        self.__zoomOutAct = viewMenu.addAction(
            getIcon('zoomout.png'), 'Zoom &out', self.em.zoomOut)
        self.__zoom11Act = viewMenu.addAction(
            getIcon('zoomreset.png'), 'Zoom r&eset', self.em.zoomReset)
        return viewMenu

    def __buildOptionsMenu(self, menuBar):
        """Build the options menu"""
        optionsMenu = QMenu("Optio&ns", menuBar)
        optionsMenu.setObjectName('options')
        optionsMenu.aboutToShow.connect(self.__optionsAboutToShow)

        self.__ideTemplateMenu = QMenu("IDE-wide &template", self)
        self.__ideCreateTemplateAct = self.__ideTemplateMenu.addAction(
            getIcon('generate.png'), '&Create')
        self.__ideCreateTemplateAct.triggered.connect(
            self._onCreateIDETemplate)
        self.__ideEditTemplateAct = self.__ideTemplateMenu.addAction(
            getIcon('edit.png'), '&Edit')
        self.__ideEditTemplateAct.triggered.connect(self._onEditIDETemplate)
        self.__ideTemplateMenu.addSeparator()
        self.__ideDelTemplateAct = self.__ideTemplateMenu.addAction(
            getIcon('trash.png'), '&Delete')
        self.__ideDelTemplateAct.triggered.connect(self._onDelIDETemplate)
        optionsMenu.addMenu(self.__ideTemplateMenu)

        optionsMenu.addSeparator()
        optionItems = [
            ('Show vertical edge', 'verticalEdge', self._verticalEdgeChanged),
            ('Show whitespaces', 'showSpaces', self._showSpacesChanged),
            ('Wrap long lines', 'lineWrap', self._lineWrapChanged),
            ('Show brace matching', 'showBraceMatch',
             self._showBraceMatchChanged),
            ('Show indentation guides', 'indentationGuides',
             self._indentationGuidesChanged),
            ('Highlight current line', 'currentLineVisible',
             self._currentLineVisibleChanged),
            ('HOME to first non-space', 'jumpToFirstNonSpace',
             self._homeToFirstNonSpaceChanged),
            ('Auto remove trailing spaces on save', 'removeTrailingOnSave',
             self._removeTrailingChanged),
            ('Editor calltips', 'editorCalltips', self._editorCalltipsChanged),
            ('Show navigation bar', 'showNavigationBar',
             self._showNavBarChanged),
            ('Show control flow navigation bar', 'showCFNavigationBar',
             self._showCFNavBarChanged),
            ('Show main toolbar', 'showMainToolBar',
             self._showMainToolbarChanged)]

        for (name, valueName, handler) in optionItems:
            act = optionsMenu.addAction(name)
            act.setCheckable(True)
            act.setChecked(self.settings[valueName])
            act.changed.connect(handler)

        optionsMenu.addSeparator()
        redirectedMenu = optionsMenu.addMenu("Redirected I/O")
        optionItems = [
            ('Reuse I/O console if available and clear', CLEAR_AND_REUSE),
            ('Reuse I/O console if available without clearing',
             NO_CLEAR_AND_REUSE),
            ('Create new I/O console for a new session', NO_REUSE)]

        self.__ioGroup = QActionGroup(self)
        for (name, data) in optionItems:
            act = redirectedMenu.addAction(name)
            act.setCheckable(True)
            act.setData(data)
            act.setActionGroup(self.__ioGroup)
            act.setChecked(self.settings['ioconsolereuse'] == data)
        redirectedMenu.triggered.connect(self._reuseIOConsoleChanged)

        tooltipsMenu = optionsMenu.addMenu("Tooltips")
        optionItems = [
            ('&Project tab', 'projectTooltips', self._projectTooltipsChanged),
            ('&Recent tab', 'recentTooltips', self._recentTooltipsChanged),
            ('&Classes tab', 'classesTooltips', self._classesTooltipsChanged),
            ('&Functions tab', 'functionsTooltips',
             self._functionsTooltipsChanged),
            ('&Outline tab', 'outlineTooltips', self._outlineTooltipsChanged),
            ('Find &name dialog', 'findNameTooltips',
             self._findNameTooltipsChanged),
            ('Find fi&le dialog', 'findFileTooltips',
             self._findFileTooltipsChanged),
            ('&Editor tabs', 'editorTooltips', self._editorTooltipsChanged)]

        for (name, valueName, handler) in optionItems:
            act = tooltipsMenu.addAction(name)
            act.setCheckable(True)
            act.setChecked(self.settings[valueName])
            act.changed.connect(handler)

        openTabsMenu = optionsMenu.addMenu("Open tabs button")
        self.__navigationSortGroup = QActionGroup(self)
        self.__alphasort = openTabsMenu.addAction("Sort alphabetically")
        self.__alphasort.setCheckable(True)
        self.__alphasort.setData(-1)
        self.__alphasort.setActionGroup(self.__navigationSortGroup)
        self.__tabsort = openTabsMenu.addAction("Tab order sort")
        self.__tabsort.setCheckable(True)
        self.__tabsort.setData(-2)
        self.__tabsort.setActionGroup(self.__navigationSortGroup)
        if self.settings['tablistsortalpha']:
            self.__alphasort.setChecked(True)
        else:
            self.__tabsort.setChecked(True)
        openTabsMenu.addSeparator()
        tabOrderPreservedAct = openTabsMenu.addAction(
            "Tab order preserved on selection")
        tabOrderPreservedAct.setCheckable(True)
        tabOrderPreservedAct.setData(0)
        tabOrderPreservedAct.setChecked(self.settings['taborderpreserved'])
        tabOrderPreservedAct.changed.connect(self._tabOrderPreservedChanged)
        openTabsMenu.triggered.connect(self._openTabsMenuTriggered)

        optionsMenu.addSeparator()
        skinsMenu = optionsMenu.addMenu("Skins")
        self.__skinsGroup = QActionGroup(self)
        currentSkin = self.settings['skin']
        for skin in getSkinsList():
            skinAct = skinsMenu.addAction(skin)
            skinAct.setData(skin)
            skinAct.setCheckable(True)
            skinAct.setActionGroup(self.__skinsGroup)
            skinAct.setChecked(currentSkin == skin)
        skinsMenu.triggered.connect(self._onSkin)

        styleMenu = optionsMenu.addMenu("Styles")
        self.__styleGroup = QActionGroup(self)
        availableStyles = QStyleFactory.keys()
        self.__styles = []
        for style in availableStyles:
            styleAct = styleMenu.addAction(style)
            styleAct.setData(style)
            styleAct.setCheckable(True)
            styleAct.setActionGroup(self.__styleGroup)
            self.__styles.append((str(style), styleAct))
        styleMenu.triggered.connect(self._onStyle)
        styleMenu.aboutToShow.connect(self.__styleAboutToShow)

        fontIcon = getIcon('fontmenu.png')
        optionsMenu.addAction(fontIcon, 'Text editor mono font...',
                              self.__onEditorFont)
        optionsMenu.addAction(fontIcon, 'Text editor margin font...',
                              self.__onEditorMarginFont)
        optionsMenu.addAction(fontIcon, 'Control flow mono font...',
                              self.__onControlFlowFont)
        optionsMenu.addAction(fontIcon, 'Control flow badge font...',
                              self.__onControlFlowBadgeFont)
        return optionsMenu

    def __buildPluginsMenu(self, menuBar):
        """Build the plugins menu"""
        self.__pluginsMenu = QMenu("P&lugins", menuBar)
        self.__pluginsMenu.setObjectName('plugins')
        self._recomposePluginMenu()
        return self.__pluginsMenu

    def __buildHelpMenu(self, menuBar):
        """Build the help menu"""
        helpMenu = QMenu("&Help", menuBar)
        helpMenu.setObjectName('help')
        helpMenu.aboutToShow.connect(self.__helpAboutToShow)
        helpMenu.aboutToHide.connect(self.__helpAboutToHide)
        self.__embeddedHelpAct = helpMenu.addAction(
            getIcon('embhelpmenu.png'),
            '&Documentation', self._onEmbeddedHelp)
        self.__shortcutsAct = helpMenu.addAction(
            getIcon('shortcutsmenu.png'),
            '&Major shortcuts', self.em.onHelp, 'F1')
        self.__contextHelpAct = helpMenu.addAction(
            getIcon('helpviewer.png'),
            'Current &word help', self._onContextHelp)
        helpMenu.addSeparator()
        self.__allShotcutsAct = helpMenu.addAction(
            getIcon('allshortcutsmenu.png'),
            '&All shortcuts (web page)', self._onAllShortcurs)
        self.__homePageAct = helpMenu.addAction(
            getIcon('homepagemenu.png'),
            'Codimension &home page', self._onHomePage)
        helpMenu.addSeparator()
        self.__aboutAct = helpMenu.addAction(
            getIcon("logo.png"), "A&bout codimension", self._onAbout)
        return helpMenu

    def __prjAboutToShow(self):
        """Triggered when project menu is about to show"""
        self.__newProjectAct.setEnabled(not self.debugMode)
        self.__openProjectAct.setEnabled(not self.debugMode)
        self._unloadProjectAct.setEnabled(not self.debugMode)

        # Recent projects part
        self.__recentPrjMenu.clear()
        addedCount = 0
        currentPrj = GlobalData().project.fileName
        for item in self.settings['recentProjects']:
            if item != currentPrj:
                addedCount += 1
                act = self.__recentPrjMenu.addAction(
                    getAccelerator(addedCount) +
                    os.path.basename(item).replace(".cdm3", ""))
                act.setData(item)
                act.setEnabled(not self.debugMode and os.path.exists(item))
        self.__recentPrjMenu.setEnabled(addedCount > 0)

        if GlobalData().project.isLoaded():
            # Template part
            exists = os.path.exists(getProjectTemplateFile())
            self._prjTemplateMenu.setEnabled(True)
            self.__createPrjTemplateAct.setEnabled(not exists)
            self.__editPrjTemplateAct.setEnabled(exists)
            self.__delPrjTemplateAct.setEnabled(exists)
        else:
            self._prjTemplateMenu.setEnabled(False)

    def __prjAboutToHide(self):
        """Triggered when project menu is about to hide"""
        self.__newProjectAct.setEnabled(True)
        self.__openProjectAct.setEnabled(True)

    def __tabAboutToShow(self):
        """Triggered when tab menu is about to show"""
        plainTextBuffer = self.__isPlainTextBuffer()
        isPythonBuffer = self._isPythonBuffer()
        isGeneratedDiagram = self.__isGeneratedDiagram()
        isProfileViewer = self.__isProfileViewer()

        self.__cloneTabAct.setEnabled(plainTextBuffer)
        self.__closeOtherTabsAct.setEnabled(self.em.closeOtherAvailable())
        self.__saveFileAct.setEnabled(
            plainTextBuffer or isGeneratedDiagram or isProfileViewer)
        self.__saveFileAsAct.setEnabled(
            plainTextBuffer or isGeneratedDiagram or isProfileViewer)
        self.__closeTabAct.setEnabled(self.em.isTabClosable())
        self.__tabJumpToDefAct.setEnabled(isPythonBuffer)
        self.__calltipAct.setEnabled(isPythonBuffer)
        self.__tabJumpToScopeBeginAct.setEnabled(isPythonBuffer)
        self.__tabOpenImportAct.setEnabled(isPythonBuffer)
        if plainTextBuffer:
            editor = self.em.currentWidget().getEditor()
            self.__openAsFileAct.setEnabled(editor.openAsFileAvailable())
            self.__downloadAndShowAct.setEnabled(
                editor.downloadAndShowAvailable())
            self.__openInBrowserAct.setEnabled(
                editor.downloadAndShowAvailable())
        else:
            self.__openAsFileAct.setEnabled(False)
            self.__downloadAndShowAct.setEnabled(False)
            self.__openInBrowserAct.setEnabled(False)

        self.__highlightInPrjAct.setEnabled(
            self.em.isHighlightInPrjAvailable())
        self.__highlightInFSAct.setEnabled(
            self.em.isHighlightInFSAvailable())
        self.__highlightInOutlineAct.setEnabled(isPythonBuffer)

        self.__closeTabAct.setShortcut("Ctrl+F4")
        self.__tabJumpToDefAct.setShortcut("Ctrl+\\")
        self.__calltipAct.setShortcut("Ctrl+/")
        self.__tabJumpToScopeBeginAct.setShortcut("Alt+U")
        self.__tabOpenImportAct.setShortcut("Ctrl+I")
        self.__highlightInOutlineAct.setShortcut("Ctrl+B")

        self.__recentFilesMenu.clear()
        addedCount = 0

        for item in getRecentFiles():
            addedCount += 1
            act = self.__recentFilesMenu.addAction(
                getAccelerator(addedCount) + item)
            act.setData(item)
            act.setEnabled(os.path.exists(item))

        self.__recentFilesMenu.setEnabled(addedCount > 0)

    def __tabAboutToHide(self):
        """Triggered when tab menu is about to hide"""
        self.__closeTabAct.setShortcut("")
        self.__tabJumpToDefAct.setShortcut("")
        self.__calltipAct.setShortcut("")
        self.__tabJumpToScopeBeginAct.setShortcut("")
        self.__tabOpenImportAct.setShortcut("")
        self.__highlightInOutlineAct.setShortcut("")
        self.__saveFileAct.setEnabled(True)
        self.__saveFileAsAct.setEnabled(True)

    def __searchAboutToShow(self):
        """Triggered when search menu is about to show"""
        isPlainTextBuffer = self.__isPlainTextBuffer()
        isPythonBuffer = self._isPythonBuffer()
        currentWidget = self.em.currentWidget()

        self.__findOccurencesAct.setEnabled(
            isPythonBuffer and os.path.isabs(currentWidget.getFileName()))
        self.__goToLineAct.setEnabled(isPlainTextBuffer)
        self.__findAct.setEnabled(isPlainTextBuffer)
        self.__replaceAct.setEnabled(
            isPlainTextBuffer and currentWidget.getType() !=
            MainWindowTabWidgetBase.VCSAnnotateViewer)
        self.__findNextAct.setEnabled(isPlainTextBuffer)
        self.__findPrevAct.setEnabled(isPlainTextBuffer)

        self.__findOccurencesAct.setShortcut("Ctrl+]")
        self.__goToLineAct.setShortcut("Ctrl+G")
        self.__findAct.setShortcut("Ctrl+F")
        self.__replaceAct.setShortcut("Ctrl+R")
        self.__findNextAct.setShortcut("Ctrl+.")
        self.__findPrevAct.setShortcut("Ctrl+,")

    def __searchAboutToHide(self):
        """Triggered when search menu is about to hide"""
        self.__findOccurencesAct.setShortcut("")
        self.__goToLineAct.setShortcut("")
        self.__findAct.setShortcut("")
        self.__replaceAct.setShortcut("")
        self.__findNextAct.setShortcut("")
        self.__findPrevAct.setShortcut("")

    def __diagramsAboutToShow(self):
        """Triggered when the diagrams menu is about to show"""
        isPythonBuffer = self._isPythonBuffer()
        widget = self.em.currentWidget()

        enabled = isPythonBuffer and \
            widget.getType() != MainWindowTabWidgetBase.VCSAnnotateViewer
        self.__tabImportDgmAct.setEnabled(enabled)
        self.__tabImportDgmDlgAct.setEnabled(enabled)

    def __runAboutToShow(self):
        """Triggered when the run menu is about to show"""
        projectLoaded = GlobalData().project.isLoaded()
        prjScriptValid = GlobalData().isProjectScriptValid()

        enabled = projectLoaded and prjScriptValid and not self.debugMode
        self.__prjRunAct.setEnabled(enabled)
        self.__prjRunDlgAct.setEnabled(enabled)

        self.__prjProfileAct.setEnabled(enabled)
        self.__prjProfileDlgAct.setEnabled(enabled)

    def __debugAboutToShow(self):
        """Triggered when the debug menu is about to show"""
        projectLoaded = GlobalData().project.isLoaded()
        prjScriptValid = GlobalData().isProjectScriptValid()

        enabled = projectLoaded and prjScriptValid and not self.debugMode
        self._prjDebugAct.setEnabled(enabled)
        self._prjDebugDlgAct.setEnabled(enabled)

    def __toolsAboutToShow(self):
        """Triggered when tools menu is about to show"""
        isPythonBuffer = self._isPythonBuffer()
        projectLoaded = GlobalData().project.isLoaded()

        self._deadCodeMenuAct.setEnabled(projectLoaded)
        self._tabDeadCodeAct.setEnabled(isPythonBuffer)
        self.disasmMenu.setEnabled(isPythonBuffer)

    def __viewAboutToShow(self):
        """Triggered when view menu is about to show"""
        isPlainTextBuffer = self.__isPlainTextBuffer()
        isGraphicsBuffer = self.__isGraphicsBuffer()
        isGeneratedDiagram = self.__isGeneratedDiagram()
        isProfileViewer = self.__isProfileViewer()
        isDiffViewer = self.__isDiffViewer()
        zoomEnabled = isPlainTextBuffer or isGraphicsBuffer or \
                      isGeneratedDiagram or isDiffViewer
        if not zoomEnabled and isProfileViewer:
            zoomEnabled = self.em.currentWidget().isZoomApplicable()
        self.__zoomInAct.setEnabled(zoomEnabled)
        self.__zoomOutAct.setEnabled(zoomEnabled)
        self.__zoom11Act.setEnabled(zoomEnabled)

        self.__zoomInAct.setShortcut("Ctrl+=")
        self.__zoomOutAct.setShortcut("Ctrl+-")
        self.__zoom11Act.setShortcut("Ctrl+0")

        self.__debugBarAct.setEnabled(self.debugMode)

    def __viewAboutToHide(self):
        """Triggered when view menu is about to hide"""
        self.__zoomInAct.setShortcut("")
        self.__zoomOutAct.setShortcut("")
        self.__zoom11Act.setShortcut("")

    def __optionsAboutToShow(self):
        """Triggered when the options menu is about to show"""
        exists = os.path.exists(getIDETemplateFile())
        self.__ideCreateTemplateAct.setEnabled(not exists)
        self.__ideEditTemplateAct.setEnabled(exists)
        self.__ideDelTemplateAct.setEnabled(exists)

    def __helpAboutToShow(self):
        """Triggered when help menu is about to show"""
        isPythonBuffer = self._isPythonBuffer()
        self.__contextHelpAct.setEnabled(isPythonBuffer)
        self.__contextHelpAct.setShortcut("Ctrl+F1")

    def __helpAboutToHide(self):
        """Triggered when help menu is about to hide"""
        self.__contextHelpAct.setShortcut("")

    def __editAboutToShow(self):
        """Triggered when edit menu is about to show"""
        isPlainBuffer = self.__isPlainTextBuffer()
        if isPlainBuffer:
            editor = self.em.currentWidget().getEditor()

        editable = isPlainBuffer and not editor.isReadOnly()
        self.__undoAct.setShortcut("Ctrl+Z")
        self.__undoAct.setEnabled(isPlainBuffer and
                                  editor.document().isUndoAvailable())
        self.__redoAct.setShortcut("Ctrl+Y")
        self.__redoAct.setEnabled(isPlainBuffer and
                                  editor.document().isRedoAvailable())
        self.__cutAct.setShortcut("Ctrl+X")
        self.__cutAct.setEnabled(editable)
        self.__copyAct.setShortcut("Ctrl+C")
        self.__copyAct.setEnabled(self.em.isCopyAvailable())
        self.__pasteAct.setShortcut("Ctrl+V")
        self.__pasteAct.setEnabled(editable and
                                   QApplication.clipboard().text() != "")
        self.__selectAllAct.setShortcut("Ctrl+A")
        self.__selectAllAct.setEnabled(isPlainBuffer)
        self.__commentAct.setShortcut("Ctrl+M")
        self.__commentAct.setEnabled(editable)
        self.__duplicateAct.setShortcut("Ctrl+D")
        self.__duplicateAct.setEnabled(editable)
        self.__autocompleteAct.setShortcut("Ctrl+Space")
        self.__autocompleteAct.setEnabled(editable)
        self.__expandTabsAct.setEnabled(editable)
        self.__trailingSpacesAct.setEnabled(editable)

    def __editAboutToHide(self):
        """Triggered when edit menu is about to hide"""
        self.__undoAct.setShortcut("")
        self.__redoAct.setShortcut("")
        self.__cutAct.setShortcut("")
        self.__copyAct.setShortcut("")
        self.__pasteAct.setShortcut("")
        self.__selectAllAct.setShortcut("")
        self.__commentAct.setShortcut("")
        self.__duplicateAct.setShortcut("")
        self.__autocompleteAct.setShortcut("")

    def __onDumpDbgSettingsAboutToShow(self):
        """Dump debug settings is about to show"""
        scriptAvailable = self._dumpScriptDbgSettingsAvailable()
        self.__debugDumpScriptSettingsAct.setEnabled(scriptAvailable)
        self.__debugDumpScriptSettingsEnvAct.setEnabled(scriptAvailable)

        projectAvailable = self._dumpProjectDbgSettingsAvailable()
        self.__debugDumpProjectSettingsAct.setEnabled(projectAvailable)
        self.__debugDumpPrjSettingsEnvAct.setEnabled(projectAvailable)

    def __styleAboutToShow(self):
        """Style menu is about to show"""
        currentStyle = self.settings['style'].lower()
        for item in self.__styles:
            item[1].setChecked(item[0].lower() == currentStyle)

    def _recomposePluginMenu(self):
        """Recomposes the plugin menu"""
        self.__pluginsMenu.clear()
        self.__pluginsMenu.addAction(getIcon('pluginmanagermenu.png'),
                                     'Plugin &manager', self._onPluginManager)
        if self._pluginMenus:
            self.__pluginsMenu.addSeparator()
        for path in self._pluginMenus:
            self.__pluginsMenu.addMenu(self._pluginMenus[path])

    def __isPlainTextBuffer(self):
        """Provides if saving is enabled"""
        currentWidget = self.em.currentWidget()
        if currentWidget is None:
            return False
        return currentWidget.getType() in \
            [MainWindowTabWidgetBase.PlainTextEditor,
             MainWindowTabWidgetBase.VCSAnnotateViewer]

    def __isGraphicsBuffer(self):
        """True if is pictures viewer"""
        currentWidget = self.em.currentWidget()
        if currentWidget is None:
            return False
        return currentWidget.getType() == MainWindowTabWidgetBase.PictureViewer

    def __isGeneratedDiagram(self):
        """True if this is a generated diagram"""
        currentWidget = self.em.currentWidget()
        if currentWidget is None:
            return False
        if currentWidget.getType() == MainWindowTabWidgetBase.GeneratedDiagram:
            return True
        if currentWidget.getType() == MainWindowTabWidgetBase.ProfileViewer:
            if currentWidget.isDiagramActive():
                return True
        return False

    def __isProfileViewer(self):
        """True if this is a profile viewer"""
        currentWidget = self.em.currentWidget()
        if currentWidget is None:
            return False
        return currentWidget.getType() == MainWindowTabWidgetBase.ProfileViewer

    def __isDiffViewer(self):
        """True if this is a diff viewer"""
        currentWidget = self.em.currentWidget()
        if currentWidget is None:
            return False
        return currentWidget.getType() == MainWindowTabWidgetBase.DiffViewer

    def _reuseIOConsoleChanged(self, act):
        """Triggered when a reuse I/O setting is changed"""
        self.settings['ioconsolereuse'] = act.data()

    def __selectFont(self, scalable, nonScalable, mono, proportional,
                     title, settingName):
        """Returns the new font or None"""
        dialog = QFontDialog(self)
        dialog.setOption(QFontDialog.DontUseNativeDialog, True)
        dialog.setOption(QFontDialog.ScalableFonts, scalable)
        dialog.setOption(QFontDialog.NonScalableFonts, nonScalable)
        dialog.setOption(QFontDialog.MonospacedFonts, mono)
        dialog.setOption(QFontDialog.ProportionalFonts, proportional)
        dialog.setWindowTitle('Select ' + title + ' for zoom level 0')
        dialog.setCurrentFont(GlobalData().skin[settingName])

        if dialog.exec_():
            font = dialog.selectedFont()
            if font.toString() != GlobalData().skin[settingName].toString():
                return font
        return None

    def __onEditorFont(self):
        """Select editor font"""
        font = self.__selectFont(False, False, True, False,
                                 'editor font', 'monoFont')
        if font is not None:
            GlobalData().skin.setTextMonoFont(font)
            self.em.onTextZoomChanged()
            self.onTextZoomChanged()

    def __onEditorMarginFont(self):
        """Select editor margin font"""
        font = self.__selectFont(True, False, True, True,
                                 'editor margin font', 'lineNumFont')
        if font is not None:
            GlobalData().skin.setMarginFont(font)
            self.em.onTextZoomChanged()

    def __onControlFlowFont(self):
        """Select the control flow mono font"""
        font = self.__selectFont(False, False, True, False,
                                 'control flow mono font', 'cfMonoFont')
        if font is not None:
            GlobalData().skin.setFlowMonoFont(font)
            self.em.onFlowZoomChanged()

    def __onControlFlowBadgeFont(self):
        """Select the control flow badge font"""
        font = self.__selectFont(True, False, True, True,
                                 'control flow badge font', 'badgeFont')
        if font is not None:
            GlobalData().skin.setFlowBadgeFont(font)
            self.em.onFlowZoomChanged()




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""custom colors dialog"""

from sys import maxsize
from ui.qt import (QDialog, QVBoxLayout, QGridLayout, QLabel, QDialogButtonBox,
                   Qt, QGraphicsRectItem, QPen, QBrush, QGraphicsScene,
                   QGraphicsView, QTimer)
from ui.colorbutton import ColorButton
from flowui.cflowsettings import getCflowSettings
from utils.pixmapcache import getIcon


class CustomColorsDialog(QDialog):

    """Custom colors dialog implementation"""

    def __init__(self, bgcolor, fgcolor, bordercolor, parent=None):
        """colors are instances of QColor"""
        QDialog.__init__(self, parent)
        self.setWindowTitle('Custom colors')

        self.__createLayout()
        self.__bgColorButton.setColor(bgcolor)
        self.__bgColorButton.sigColorChanged.connect(self.__onColorChanged)
        self.__fgColorButton.setColor(fgcolor)
        self.__fgColorButton.sigColorChanged.connect(self.__onColorChanged)
        if bordercolor is None:
            self.__borderColorButton.setEnabled(False)
            self.__borderColorButton.setToolTip(
                'Border colors are not supported for docstrings')
            self.__borderColorButton.setIcon(getIcon('warning.png'))
        else:
            self.__borderColorButton.setColor(bordercolor)
            self.__borderColorButton.sigColorChanged.connect(
                self.__onColorChanged)

        QTimer.singleShot(1, self.__onColorChanged)

    def __createLayout(self):
        """Creates the dialog layout"""
        self.setMinimumWidth(300)
        self.setMinimumHeight(200)
        self.resize(300, 200)
        self.setSizeGripEnabled(True)

        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(5, 5, 5, 5)
        gridLayout = QGridLayout()

        bgLabel = QLabel('Select background color:', self)
        gridLayout.addWidget(bgLabel, 0, 0, 1, 1)
        self.__bgColorButton = ColorButton('', self)
        gridLayout.addWidget(self.__bgColorButton, 0, 1, 1, 1)

        fgLabel = QLabel('Select foreground color:', self)
        gridLayout.addWidget(fgLabel, 1, 0, 1, 1)
        self.__fgColorButton = ColorButton('', self)
        gridLayout.addWidget(self.__fgColorButton, 1, 1, 1, 1)

        borderLabel = QLabel('Select border color:', self)
        gridLayout.addWidget(borderLabel, 2, 0, 1, 1)
        self.__borderColorButton = ColorButton('', self)
        gridLayout.addWidget(self.__borderColorButton, 2, 1, 1, 1)

        verticalLayout.addLayout(gridLayout)

        # Sample area
        self.__scene = QGraphicsScene()
        self.__view = QGraphicsView()
        self.__view.setScene(self.__scene)
        verticalLayout.addWidget(self.__view)

        # Buttons at the bottom
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Cancel |
                                     QDialogButtonBox.Ok)
        verticalLayout.addWidget(buttonBox)

        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

    def __onColorChanged(self):
        """The user changed the color so redraw the sample"""
        viewWidth = self.__view.width()
        viewHeight = self.__view.height()

        self.__scene.clear()
        # without '-4' scrollbar will appear
        self.__scene.setSceneRect(0, 0, viewWidth - 4, viewHeight - 4)
        block = SampleBlock(getCflowSettings(self),
                            self.backgroundColor(),
                            self.foregroundColor(),
                            self.borderColor(),
                            viewWidth, viewHeight)
        self.__scene.addItem(block)
        self.__scene.update()

    def backgroundColor(self):
        """Provides the background color"""
        return self.__bgColorButton.color()

    def foregroundColor(self):
        """Provides the foreground color"""
        return self.__fgColorButton.color()

    def borderColor(self):
        """Provides the border color"""
        return self.__borderColorButton.color()


class SampleBlock(QGraphicsRectItem):

    """Sample block"""

    def __init__(self, settings, bgColor, fgColor, borderColor, width, height):
        QGraphicsRectItem.__init__(self)
        self.__settings = settings
        self.__bgColor = bgColor
        self.__fgColor = fgColor
        self.__borderColor = borderColor
        self.__viewWidth = width
        self.__viewHeight = height

        self.__textRect = self.__settings.monoFontMetrics.boundingRect(
            0, 0, maxsize, maxsize, 0, 'Sample')


        vPadding = 2 * settings.vTextPadding
        self.__rectHeight = self.__textRect.height() + vPadding
        hPadding = 2 * settings.hTextPadding
        self.__rectWidth = max(self.__textRect.width() + hPadding,
                               settings.minWidth)

    def paint(self, painter, option, widget):
        """Draws the code block"""
        del option
        del widget

        baseX = (self.__viewWidth - self.__rectWidth) / 2
        baseY = (self.__viewHeight - self.__rectHeight) / 2

        if self.__borderColor is None:
            pen = QPen(self.__bgColor)
        else:
            pen = QPen(self.__borderColor)
        painter.setPen(pen)
        brush = QBrush(self.__bgColor)
        painter.setBrush(brush)
        painter.drawRect(baseX, baseY, self.__rectWidth, self.__rectHeight)

        # Draw the text in the rectangle
        pen = QPen(self.__fgColor)
        painter.setFont(self.__settings.monoFont)
        painter.setPen(pen)

        textWidth = self.__textRect.width() + 2 * self.__settings.hTextPadding
        textShift = (self.__rectWidth - textWidth) / 2
        painter.drawText(baseX + self.__settings.hTextPadding + textShift,
                         baseY + self.__settings.vTextPadding,
                         self.__textRect.width(), self.__textRect.height(),
                         Qt.AlignLeft, 'Sample')



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Control flow UI graphics scene and a view"""

from ui.qt import (Qt, QRectF, QPoint, QPainter, QGraphicsView, QApplication,
                   QGraphicsScene)
from utils.settings import Settings
from .flowuicontextmenus import CFSceneContextMenuMixin
from .flowuimouse import CFSceneMouseMixin
from .flowuikeyboard import CFSceneKeyboardMixin



class CFGraphicsScene(QGraphicsScene,
                      CFSceneContextMenuMixin,
                      CFSceneMouseMixin,
                      CFSceneKeyboardMixin):

    """Reimplemented graphics scene"""

    def __init__(self, navBar, parent=None):
        QGraphicsScene.__init__(self, parent)
        CFSceneContextMenuMixin.__init__(self)
        CFSceneMouseMixin.__init__(self)
        CFSceneKeyboardMixin.__init__(self)
        self.__navBar = navBar
        self.selectionChanged.connect(self.selChanged)

    def selChanged(self):
        """Triggered when a selection changed"""
        items = self.sortSelectedReverse()
        items.reverse()
        count = len(items)
        if count:
            tooltip = []
            for item in items:
                if hasattr(item, "getSelectTooltip"):
                    tooltip.append(item.getSelectTooltip())
                else:
                    tooltip.append(str(type(item)))
            self.__navBar.setSelectionLabel(count, "\n".join(tooltip))
        else:
            self.__navBar.setSelectionLabel(0, None)

    def serializeSelection(self):
        """Builds a list of pairs: (id->int, tooltip->str)"""
        selection = []
        for item in self.selectedItems():
            selection.append((item.itemID, item.getSelectTooltip()))
        return selection

    # The selection can be restored by item IDs if the number of items on the
    # diagram has not changed. E.g. when zoom is done or a cml comment is
    # added/removed
    def restoreSelectionByID(self, selection):
        """Restores the selection by the item ID"""
        if selection:
            ids = [item[0] for item in selection]
            for item in self.items():
                if hasattr(item, "itemID"):
                    if item.itemID in ids:
                        item.setSelected(True)

    # The selection can be restored by item tooltips if there is no text
    # modifications but there are representation changes. E.g. the selection
    # can be roughly preserved between the smart zoom levels or when except
    # blocks are hidden/shown
    def restoreSelectionByTooltip(self, selection):
        """Restores the selection by the item tooltip"""
        if selection:
            tooltips = [item[1] for item in selection]
            for item in self.items():
                if hasattr(item, "getSelectTooltip"):
                    if item.getSelectTooltip() in tooltips:
                        item.setSelected(True)

    def getDocItemByAnchor(self, anchor):
        """Provides the graphics item for the given anchor if so"""
        for item in self.items():
            if item.isCMLDoc():
                if item.cmlRef.anchor == anchor:
                    return item
        return None

    def terminate(self):
        """Called when a tab is closed"""
        self.selectionChanged.disconnect(self.selChanged)
        self.clear()
        self.terminateMenus()



class CFGraphicsView(QGraphicsView):

    """Central widget"""

    def __init__(self, navBar, parent):
        super(CFGraphicsView, self).__init__(parent)
        self.scene = CFGraphicsScene(navBar, parent)
        self.setScene(self.scene)

        self.__parentWidget = parent
        self.__currentFactor = 1.0
        self.setRenderHint(QPainter.Antialiasing)
        self.setRenderHint(QPainter.TextAntialiasing)

    def wheelEvent(self, event):
        """Mouse wheel event"""
        modifiers = QApplication.keyboardModifiers()
        if modifiers == Qt.ControlModifier:
            angleDelta = event.angleDelta()
            if not angleDelta.isNull():
                if angleDelta.y() > 0:
                    Settings().onFlowZoomIn()
                else:
                    Settings().onFlowZoomOut()
            event.accept()
        elif modifiers == Qt.ShiftModifier:
            angleDelta = event.angleDelta()
            if not angleDelta.isNull():
                if angleDelta.y() > 0:
                    self.__parentWidget.onSmartZoomLevelUp()
                else:
                    self.__parentWidget.onSmartZoomLevelDown()
            event.accept()
        else:
            QGraphicsView.wheelEvent(self, event)

    def getVisibleRect(self):
        """Provides the visible rectangle"""
        topLeft = self.mapToScene(QPoint(0, 0))
        bottomRight = self.mapToScene(QPoint(self.viewport().width(),
                                             self.viewport().height()))
        return QRectF(topLeft, bottomRight)

    def scrollTo(self, item, makeFirstOnScreen=False):
        """Scrolls the view to the item"""
        if item is None:
            return

        # When there is a change on the diagram, e.g. when a font was changed,
        # when a smart zoom was changed or a display mode was changed then
        # the sync between the views is done basing on what is the first
        # visible item. So the vertical scroll is required to make the item of
        # interest first on the screen

        visibleRect = self.getVisibleRect()
        itemRect = item.boundingRect()
        if not makeFirstOnScreen:
            if visibleRect.contains(itemRect):
                # The item is fully visible
                return

            # The item is fully visible vertically
            if itemRect.topLeft().y() >= visibleRect.topLeft().y() and \
               itemRect.bottomLeft().y() <= visibleRect.bottomLeft().y():
                self.__hScrollToItem(item)
                return

        # The item top left is visible
        if visibleRect.contains(itemRect.topLeft()):
            # So far scroll the view vertically anyway
            val = float(itemRect.topLeft().y() - 15.0)
            self.verticalScrollBar().setValue(val)
            self.__hScrollToItem(item)
            return

        # Here: the top left is not visible, so the vertical scrolling is
        # required
        val = float(itemRect.topLeft().y() - 15.0)
        self.verticalScrollBar().setValue(val)
        self.__hScrollToItem(item)

    def __hScrollToItem(self, item):
        """Sets the horizontal scrolling for the item"""
        if item is None:
            return

        # There are a few cases here:
        # - the item does not fit the screen width
        # - the item fits the screen width and would be visible if the
        #   scroll is set to 0
        # - the item fits the screen width and would not be visible if the
        #   scroll is set to 0

        visibleRect = self.getVisibleRect()
        itemRect = item.boundingRect()

        if itemRect.width() > visibleRect.width():
            # Does not fit the screen
            val = float(itemRect.topLeft().x()) - 15.0
            self.horizontalScrollBar().setValue(val)
        else:
            if itemRect.topRight().x() < visibleRect.width():
                # Fits the screen if the scroll is 0
                self.horizontalScrollBar().setValue(0)
            else:
                val = float(itemRect.topLeft().x()) - 15.0
                self.horizontalScrollBar().setValue(val)

    def terminate(self):
        """Called when a tab is closed"""
        self.scene.terminate()
        self.scene.deleteLater()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2011-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The implementation vastly derived from eric4. Here is the original copyright:
# Copyright (c) 2002 - 2013 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Debugger variable browser items"""

from ui.qt import Qt, QTreeWidgetItem
from utils.pixmapcache import getIcon

INDICATORS = ("()", "[]", "{:}", "{}")


def getDisplayValue(displayValue):
    """Takes potentially multilined value and converts it to a single line"""
    lines = str(displayValue).splitlines()
    lineCount = len(lines)
    if lineCount > 1:
        # There are many lines. Find first non-empty.
        nonEmptyIndex = None
        index = -1
        for line in lines:
            index += 1
            if len(line.strip()) > 0:
                nonEmptyIndex = index
                break
        if nonEmptyIndex is None:
            displayValue = ""   # Multilined empty string
        else:
            if len(lines[nonEmptyIndex]) > 128:
                displayValue = lines[nonEmptyIndex][:128] + "<...>"
            else:
                displayValue = lines[nonEmptyIndex]
                if nonEmptyIndex < lineCount - 1:
                    displayValue += "<...>"

            if nonEmptyIndex > 0:
                displayValue = "<...>" + displayValue
    elif lineCount == 1:
        # There is just one line
        if len(lines[0]) > 128:
            displayValue = lines[0][:128] + "<...>"
        else:
            value = lines[0]
    return displayValue


def getTooltipValue(value):
    """Takes a potentially multilined string and converts it to
       the form suitable for tooltips
    """
    lines = value.splitlines()
    lineCount = len(lines)
    if lineCount > 1:
        value = ""
        index = 0
        for line in lines:
            if index >= 5:  # First 5 lines only
                break
            if index > 0:
                value += "\n"
            if len(line) > 128:
                value += line[:128] + "<...>"
            else:
                value += line
            index += 1
        if lineCount > 5:
            value += "\n<...>"
    elif lineCount == 1:
        if len(lines[0]) > 128:
            value = lines[0][:128] + "<...>"
        else:
            value = lines[0]
    return value


class VariableItem(QTreeWidgetItem):

    """Base structure for variable items"""

    def __init__(self, parent, isGlobal,
                 displayName, displayValue, displayType):
        self.__isGlobal = isGlobal
        self.__value = displayValue
        self.__type = displayType

        self.__name, self.__varID = VariableItem.extractId(displayName)

        # Decide about the display value
        displayValue = getDisplayValue(displayValue)

        # Decide about the tooltip
        self.__tooltip = "Name: " + self.__name + "\n" + \
                         "Type: " + displayType + "\n" + \
                         "Value: "

        tooltipDisplayValue = getTooltipValue(self.__value)
        if '\r' in tooltipDisplayValue or '\n' in tooltipDisplayValue:
            self.__tooltip += "\n" + tooltipDisplayValue
        else:
            self.__tooltip += tooltipDisplayValue

        QTreeWidgetItem.__init__(self, parent, [self.__name, displayValue,
                                                displayType])
        self.populated = True

    def getValue(self):
        """Provides the variable value"""
        return self.__value

    def getId(self):
        """Provides the variable ID"""
        return self.__varID

    def getName(self):
        """Provides the variable name"""
        return self.__name

    def getType(self):
        """Provides the variable type"""
        return self.__type

    def isGlobal(self):
        """Tells if the variable is global"""
        return self.__isGlobal

    @classmethod
    def extractId(cls, var):
        """Extracts the variable ID"""
        if " (ID:" in var:
            dvar, varID = var.rsplit(None, 1)
            if varID.endswith(INDICATORS):
                varID, indicators = VariableItem.extractIndicators(varID)
                dvar += indicators
        else:
            dvar = var
            varID = None
        return dvar, varID

    @classmethod
    def extractIndicators(cls, var):
        """Extract the indicator string from a variable text"""
        for indicator in INDICATORS:
            if var.endswith(indicator):
                return var[:-len(indicator)], indicator
        return var, ""

    def _buildKey(self):
        """Builds a key to access to the variable"""
        indicators = ""
        txt = self.text(0)
        if txt.endswith(INDICATORS):
            txt, indicators = VariableItem.extractIndicators(txt)
        if self.__varID:
            txt = "{0} {1}{2}".format(txt, self.__varID, indicators)
        else:
            txt = "{0}{1}".format(txt, indicators)
        return txt

    def data(self, column, role):
        """Provides the data for the requested role"""
        if role == Qt.ToolTipRole:
            return self.__tooltip
        if role == Qt.DecorationRole:
            if column == 0:
                if not self.parent():
                    if self.__isGlobal:
                        fileName = 'globvar.png'
                    else:
                        fileName = 'locvar.png'
                    return getIcon(fileName)
        return QTreeWidgetItem.data(self, column, role)

    def attachDummy(self):
        """Attach a dummy sub item to allow for lazy population"""
        QTreeWidgetItem(self, ["DUMMY"])

    def deleteChildren(self):
        """Deletes all children (cleaning the subtree)"""
        for item in self.takeChildren():
            del item

    def expand(self):
        """Does nothing for the basic item. Should be overwritten"""
        pass

    def collapse(self):
        """Does nothing for the basic item. Should be overwritten"""
        pass



class SpecialVariableItem(VariableItem):

    """These special variable nodes are generated for classes, lists,
       tuples and dictionaries.
    """

    def __init__(self, parent, debugger, isGlobal,
                 displayName, displayValue, displayType, frameNumber):
        VariableItem.__init__(self, parent, isGlobal,
                              displayName, displayValue, displayType)
        self.attachDummy()
        self.populated = False

        self.frameNumber = frameNumber
        self.__debugger = debugger

    def expand(self):
        """Expands the item"""
        self.deleteChildren()
        self.populated = True

        pathlist = [self._buildKey()]
        par = self.parent()

        # Step 1: get a pathlist up to the requested variable
        while par is not None:
            pathlist.insert(0, par._buildKey())
            par = par.parent()

        # Step 2: request the variable from the debugger
        self.__debugger.remoteClientVariable(self.isGlobal(),
                                             pathlist, self.frameNumber)


class ArrayElementVariableItem(VariableItem):

    """Represents an array element"""

    def __init__(self, parent, isGlobal,
                 displayName, displayValue, displayType):
        VariableItem.__init__(self, parent, isGlobal,
                              displayName, displayValue, displayType)

        """
        Array elements have numbers as names, but the key must be
        right justified and zero filled to 6 decimal places. Then
        element 2 will have a key of '000002' and appear before
        element 10 with a key of '000010'
        """
        col0Str = self.text(0)
        self.setText(0, "{0:6d}".format(int(col0Str)))


class SpecialArrayElementVariableItem(SpecialVariableItem):

    """Represents a special array variable node"""

    def __init__(self, parent, debugger, isGlobal,
                 displayName, displayValue, displayType, frameNumber):
        SpecialVariableItem.__init__(self, parent, debugger, isGlobal,
                                     displayName, displayValue, displayType,
                                     frameNumber)
        """
        Array elements have numbers as names, but the key must be
        right justified and zero filled to 6 decimal places. Then
        element 2 will have a key of '000002' and appear before
        element 10 with a key of '000010'
        """
        col0Str, indicators = VariableItem.extractIndicators(self.text(0))
        self.setText(0, "{0:6d}{1}".format(int(col0Str), indicators))



#
# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The code is taken from here and slightly tweaked:
# http://pydev.blogspot.ru/2014/03/should-python-garbage-collector-be.html
#


"""Custom garbage collector"""


import gc
import logging
from .qt import QObject, QTimer


class GarbageCollector(QObject):

    '''Disable automatic garbage collection and instead collect manually
       every INTERVAL milliseconds.

       This is done to ensure that garbage collection only happens in the GUI
       thread, as otherwise Qt can crash.
    '''

    INTERVAL = 10000

    def __init__(self, parent):
        QObject.__init__(self, parent)

        self.timer = QTimer(self)
        self.timer.timeout.connect(self.check)

        self.threshold = gc.get_threshold()
        gc.disable()
        self.timer.start(self.INTERVAL)

    def check(self):
        """Called by the QTimer periodically in the GUI thread"""
        # return self.debug_cycles() # uncomment to just debug cycles
        lvl0, lvl1, lvl2 = gc.get_count()
        logging.debug("gc_check called: %d, %d, %d", lvl0, lvl1, lvl2)
        if lvl0 > self.threshold[0]:
            num = gc.collect(0)
            logging.debug("collecting gen 0, found: %d unreachable", num)
            if lvl1 > self.threshold[1]:
                num = gc.collect(1)
                logging.debug("collecting gen 1, found: %d unreachable", num)
                if lvl2 > self.threshold[2]:
                    num = gc.collect(2)
                    logging.debug("collecting gen 2, found: %d unreachable",
                                  num)

    def debug_cycles(self):
        """Debugging support"""
        gc.set_debug(gc.DEBUG_SAVEALL)
        gc.collect()
        for obj in gc.garbage:
            logging.debug(repr(obj) + " " + str(type(obj)))



#
# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""imports/definitions selection widget"""

import os.path
from utils.globals import GlobalData
from utils.fileutils import isPythonFile
from .qt import (Qt, QEventLoop, QSizePolicy, QFrame, QTreeWidget,
                 QApplication, QTreeWidgetItem, QHeaderView, QVBoxLayout,
                 QAbstractItemView)
from .itemdelegates import NoOutlineHeightDelegate


class ImportsList(QTreeWidget):

    """Need to derive for overloading focusOutEvent()"""

    def __init__(self, parent):
        QTreeWidget.__init__(self, parent)

    def focusOutEvent(self, _):
        """Hides the imports list widget"""
        self.parent().hide()


class ImportListWidget(QFrame):

    """Frameless dialogue to select an import to open"""

    IMPORT_MODE = 0
    DEFINITION_MODE = 1

    def __init__(self, parent=None):
        QFrame.__init__(self, parent)

        # Make the frame nice looking
        self.setFrameShape(QFrame.Panel)
        self.setLineWidth(1)

        self.__importList = None
        self.__mode = None
        self.__createLayout()

    def __createLayout(self):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)

        self.__importList = ImportsList(self)
        self.__importList.setAlternatingRowColors(True)
        self.__importList.setRootIsDecorated(False)
        self.__importList.setItemsExpandable(False)
        self.__importList.setUniformRowHeights(True)
        self.__importList.setSelectionMode(QAbstractItemView.SingleSelection)
        self.__importList.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.__importList.setItemDelegate(NoOutlineHeightDelegate(4))

        self.__importList.itemActivated.connect(self.__importActivated)

        verticalLayout.addWidget(self.__importList)
        verticalLayout.setContentsMargins(0, 0, 0, 0)

        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizeHint = self.__importList.minimumSizeHint()
        self.setFixedHeight(sizeHint.height())
        self.move(5, 5)

    def showResolvedImports(self, importsList):
        """Pops up the dialogue"""
        self.__importList.clear()
        self.__populateImports(importsList)
        self.__mode = ImportListWidget.IMPORT_MODE

        self.__importList.setCurrentItem(self.__importList.topLevelItem(0))

        QApplication.processEvents(QEventLoop.ExcludeUserInputEvents)

        self.resize()
        self.show()
        self.__importList.setFocus()

    def resize(self):
        """Resizes the dialogue to match the editor size"""
        scroll = self.parent().verticalScrollBar()
        scrollWidth = scroll.width()

        width = self.parent().width()
        widgetWidth = width - scrollWidth - 10

        self.setFixedWidth(widgetWidth)

    def __populateImports(self, importsList):
        """Populates the dialogue with imports"""
        count = len(importsList)
        info = str(count) + " resolved import"
        if count > 1:
            info += "s"
        headerItem = QTreeWidgetItem(["Import (" + info + ")", "Path"])
        self.__importList.setHeaderItem(headerItem)
        for item in importsList:
            importName = item[0]
            resolvedPath = item[1]
            if resolvedPath is None:
                resolvedPath = ''
            importItem = QTreeWidgetItem([importName, resolvedPath])
            importItem.setToolTip(0, self.__getFileTooltip(resolvedPath))
            self.__importList.addTopLevelItem(importItem)

        self.__importList.header().resizeSections(
            QHeaderView.ResizeToContents)

    def showDefinitions(self, definitions):
        """Pops up the dialog"""
        self.__importList.clear()
        self.__populateDefinitions(definitions)
        self.__mode = ImportListWidget.DEFINITION_MODE

        self.__importList.setCurrentItem(self.__importList.topLevelItem(0))

        QApplication.processEvents(QEventLoop.ExcludeUserInputEvents)

        self.resize()
        self.show()
        self.__importList.setFocus()

    def __populateDefinitions(self, definitions):
        """Populates the dialogue with definitions"""
        headerItem = QTreeWidgetItem(["Type", "Path", "Line", "Column"])
        self.__importList.setHeaderItem(headerItem)
        for item in definitions:
            defItem = QTreeWidgetItem([item[3], item[0],
                                       str(item[1]), str(item[2] + 1)])
            self.__importList.addTopLevelItem(defItem)

        self.__importList.header().resizeSections(
            QHeaderView.ResizeToContents)

    @staticmethod
    def __getFileTooltip(path):
        """Provides the python file docstring for a tooltip"""
        path = os.path.normpath(path)
        if os.path.exists(path):
            if isPythonFile(path):
                modInfo = GlobalData().briefModinfoCache.get(path)
                if modInfo.docstring is not None:
                    return modInfo.docstring.text
        return ''

    def setFocus(self):
        """Sets the focus to the list of imports"""
        self.__importList.setFocus()

    def keyPressEvent(self, event):
        """Handles the key press events"""
        if event.key() == Qt.Key_Escape:
            editorsManager = GlobalData().mainWindow.editorsManager()
            activeWindow = editorsManager.currentWidget()
            if activeWindow:
                activeWindow.setFocus()
            event.accept()
            self.hide()

    def __importActivated(self, item, _):
        """Handles the import selection"""
        path = str(item.text(1))
        if self.__mode == ImportListWidget.IMPORT_MODE:
            if os.path.exists(path):
                GlobalData().mainWindow.openFile(path, -1)
        else:
            line = int(item.text(2))
            column = int(item.text(3))
            GlobalData().mainWindow.openFile(path, line, column)
